import os
from collections import namedtuple

from fsgs.platforms import PLATFORM_AMIGA


class KnownFilePath:
    def __init__(self, path=None):
        self.__path = path

    @property
    def path(self):
        return self.__path

    def exists(self):
        return os.path.exists(self.path)


KnownFile = namedtuple("KnownFile", ["sha1", "platform", "name"])
KnownFileMod = namedtuple("KnownFileMod", ["sha1", "mod"])


ACTION_REPLAY_MK_II_2_14_ROM = KnownFile(
    "255d6df63a4eab0a838eb1a16a267b0959dff634",
    PLATFORM_AMIGA,
    "Action Replay Mk II v2.14.rom",
)
ACTION_REPLAY_MK_II_2_14_MOD_ROM = KnownFile(
    "14b1f5a69efb6f4e2331970e6ca0f33c0f04ac91",
    PLATFORM_AMIGA,
    "Action Replay Mk II v2.14 (Mod).rom",
)
# noinspection SpellCheckingInspection
ACTION_REPLAY_MK_III_3_17_ROM = KnownFile(
    "5d4987c2e3ffea8b1b02e31430ef190f2db76542",
    PLATFORM_AMIGA,
    "Action Replay Mk III v3.17.rom",
)
ACTION_REPLAY_MK_III_3_17_MOD_ROM = KnownFile(
    "0439d6ccc2a0e5c2e83fcf2389dc4d4a440a4c62",
    PLATFORM_AMIGA,
    "Action Replay Mk III v3.17 (Mod).rom",
)

# known_files = [
#     ACTION_REPLAY_MK_III_3_17_ROM
# ]

# noinspection SpellCheckingInspection
# known_file_mods = {
#     ACTION_REPLAY_MK_III_3_17_MOD_ROM.sha1: KnownFileMod(
#         ACTION_REPLAY_MK_III_3_17_ROM.sha1, "zero-initial-4-bytes"),
# }



from __future__ import absolute_import
from __future__ import print_function

import time

ts_empty_string = "--.--.---- --:--:--.--"
ts_format = "%d.%m.%Y %H:%M:%S"

class TimeStamp:
  def __init__(self, days=0, mins=0, ticks=0):
    self.days = days
    self.mins = mins
    self.ticks = ticks
    self.secs = days * 24 * 60 * 60 + mins * 60 + (ticks / 50)
    self.sub_secs = (ticks % 50)
  
  def __str__(self):
    t = time.localtime(self.secs)
    ts = time.strftime(ts_format, t)
    return "%s t%02d" % (ts, self.sub_secs)
    
  def get_secsf(self):
    return self.secs + self.sub_secs / 50.0
  
  def get_secs(self):
    return self.secs
  
  def from_secs(self, secs):
    ticks = secs * 50
    mins = ticks / (50 * 60)
    self.ticks = ticks % (50 * 60)
    self.days = mins / (60 * 24)
    self.mins = mins % (60 * 24)
    self.secs = secs
    self.sub_secs = 0
  
  def parse(self, s):
    # check for ticks
    s = s.strip()
    ticks = 0
    if len(s) > 3:
      t = s[-3:]
      if t[0] == 't' and t[1:].isdigit():
        ticks = int(t[1:])
        s = s[:-4]
    # parse normal time
    try:
      ts = time.strptime(s, ts_format)
      secs = time.mktime(ts)
      self.from_secs(secs)
      self.sub_secs = ticks
      self.ticks += ticks
      return True
    except ValueError:
      return False
  
if __name__ == '__main__':
  ts = TimeStamp()
  ts.from_secs(123)
  ts2 = TimeStamp(days=ts.days, mins=ts.mins, ticks=ts.ticks)
  if ts2.get_secs() != 123:
    print("FAIL")
  
  ts = TimeStamp()
  s = "05.01.2012 21:47:34 t40"
  ts.parse(s)
  txt = str(ts)
  if s != txt:
    print("FAIL")



import ctypes

from OpenGL import GL as gl

from arcade.glui.opengl import fs_emu_texturing
from arcade.glui.render import Render
from arcade.glui.window import (
    premultiplied_texture_program,
    set_program,
    texture_program,
)


class FrameBufferObject(object):  # Renderable, RendererBase):
    def __init__(self, width=100, height=100):
        self._transparent = True
        self._has_transparency = True
        self._width = width
        self._height = height
        self._fb = None
        self._depth_rb = None
        self._stencil_rb = None
        self._texture = None

    def __del__(self):
        self.free()

    def free(self):
        if self._fb:
            gl.glDeleteFramebuffers(1, self._fb)
        if self._depth_rb:
            gl.glDeleteRenderbuffers(1, self._depth_rb)
        if self._stencil_rb:
            gl.glDeleteRenderbuffers(1, self._stencil_rb)
        if self._texture:
            gl.glDeleteTextures([self._texture])

    def _create(self):
        if self._fb is not None:
            return

        w = self._width
        h = self._height
        if self._transparent:
            texture_format = gl.GL_RGBA
        else:
            texture_format = gl.GL_RGBA
        texture_target = gl.GL_TEXTURE_2D
        fb = gl.GLuint()
        depth_rb = gl.GLuint()
        stencil_rb = gl.GLuint()

        gl.glGenFramebuffers(1, ctypes.byref(fb))
        color_tex = Render.get().create_texture()
        gl.glGenRenderbuffers(1, ctypes.byref(depth_rb))
        gl.glGenRenderbuffers(1, ctypes.byref(stencil_rb))

        fb = fb.value
        gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, fb)

        # initialize color texture
        gl.glBindTexture(texture_target, color_tex)
        gl.glTexParameterf(
            texture_target, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR
        )
        gl.glTexParameterf(
            texture_target, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR
        )
        gl.glTexImage2D(
            texture_target,
            0,
            texture_format,
            w,
            h,
            0,
            gl.GL_RGB,
            gl.GL_INT,
            None,
        )
        try:
            # FIXME: An error seems to be thrown here on Windows, even
            # though it seems to work...
            gl.glFramebufferTexture2D(
                gl.GL_FRAMEBUFFER,
                gl.GL_COLOR_ATTACHMENT0,
                texture_target,
                color_tex,
                0,
            )
        except Exception:
            pass

        # initialize depth renderbuffer
        gl.glBindRenderbuffer(gl.GL_RENDERBUFFER, depth_rb)
        gl.glRenderbufferStorage(
            gl.GL_RENDERBUFFER, gl.GL_DEPTH_COMPONENT24, w, h
        )
        gl.glFramebufferRenderbuffer(
            gl.GL_FRAMEBUFFER,
            gl.GL_DEPTH_ATTACHMENT,
            gl.GL_RENDERBUFFER,
            depth_rb,
        )

        # initialize stencil renderbuffer

        # glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, stencil_rb)
        # glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_STENCIL_INDEX,
        #         512, 512)
        # glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
        #        GL_STENCIL_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, stencil_rb)
        gl.glBindTexture(texture_target, 0)

        # Check framebuffer completeness at the end of initialization.
        status = gl.glCheckFramebufferStatus(gl.GL_FRAMEBUFFER)
        assert status == gl.GL_FRAMEBUFFER_COMPLETE, str(status)
        self._fb = fb
        self._depth_rb = depth_rb
        self._stencil_rb = stencil_rb
        self._texture = color_tex

    def bind(self):
        if self._fb is None:
            self._create()
        gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, self._fb)
        # glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, self._depth_rb)

    def unbind(self):
        gl.glBindFramebuffer(gl.GL_FRAMEBUFFER, 0)
        # glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0)

    def set_viewport(self):
        gl.glViewport(0, 0, self._width, self._height)

    def render(self, opacity=1.0):
        render_texture(
            self._texture,
            -1.0,
            -1.0,
            0,
            2.0,
            2.0,
            opacity=opacity,
            is_premultiplied=True,
        )


def render_texture(
    texture, x, y, z, w, h, opacity=1.0, is_premultiplied=False
):
    if is_premultiplied:
        set_program(premultiplied_texture_program)
    else:
        set_program(texture_program)
    fs_emu_texturing(True)
    gl.glBindTexture(gl.GL_TEXTURE_2D, texture)
    gl.glBegin(gl.GL_QUADS)
    gl.glColor(1.0, 1.0, 1.0, opacity)
    gl.glTexCoord2f(0.0, 0.0)
    gl.glVertex3f(x, y, z)
    gl.glTexCoord2f(1.0, 0.0)
    gl.glVertex3f(x + w, y, z)
    gl.glTexCoord2f(1.0, 1.0)
    gl.glVertex3f(x + w, y + h, z)
    gl.glTexCoord2f(0.0, 1.0)
    gl.glVertex3f(x, y + h, z)
    gl.glEnd()
    set_program(None)



from ...launcher_config import LauncherConfig
from .StatusElement import StatusElement


class LanguageElement(StatusElement):
    def __init__(self, parent, language, icon, tool_tip=""):
        StatusElement.__init__(self, parent)
        self.set_min_width(32)
        self.icon = icon
        self.language = language
        self.languages = ""
        self.hide()
        if tool_tip:
            self.set_tooltip(tool_tip)

        LauncherConfig.add_listener(self)
        self.on_config("languages", LauncherConfig.get("languages"))

    def on_destroy(self):
        LauncherConfig.remove_listener(self)

    def on_config(self, key, value):
        if key == "languages":
            if value != self.languages:
                self.languages = value
                if self.language in value:
                    self.show()
                else:
                    self.hide()



import fsui


class SetupWizardPage(fsui.Panel):
    WIDTH = 640
    HEIGHT = 400

    def __init__(self, parent):
        super().__init__(parent)
        self.set_min_width(SetupWizardPage.WIDTH)
        self.set_min_height(SetupWizardPage.HEIGHT)



from __future__ import absolute_import
from __future__ import print_function

from amitools.fs.block.Block import *
import amitools.fs.DosType as DosType

class FSHeaderDeviceNode:
  valid_flags = ('type', 'task', 'lock', 'handler', 'stack_size',
                 'priority', 'startup', 'seg_list_blk', 'global_vec')

  def __init__(self, type=0, task=0, lock=0, handler=0, stack_size=0, priority=0,
               startup=0, seg_list_blk=0, global_vec=0):
    self.type = type
    self.task = task
    self.lock = lock
    self.handler = handler
    self.stack_size = stack_size
    self.priority = priority
    self.startup = startup
    self.seg_list_blk = seg_list_blk
    self.global_vec = global_vec

  def dump(self):
    print("DeviceNode")
    print(" type:           0x%08x" % self.type)
    print(" task:           0x%08x" % self.task)
    print(" lock:           0x%08x" % self.lock)
    print(" handler:        0x%08x" % self.handler)
    print(" stack_size:     0x%08x" % self.stack_size)
    print(" seg_list_blk:   0x%08x" % self.seg_list_blk)
    print(" global_vec:     0x%08x" % self.global_vec)

  def get_flags(self, patch_flags = 0x1ff):
    res = []
    if patch_flags & 0x01 == 0x01:
      res.append(('type',self.type))
    if patch_flags & 0x02 == 0x02:
      res.append(('task',self.task))
    if patch_flags & 0x04 == 0x04:
      res.append(('lock',self.lock))
    if patch_flags & 0x08 == 0x08:
      res.append(('handler',self.handler))
    if patch_flags & 0x10 == 0x10:
      res.append(('stack_size',self.stack_size))
    if patch_flags & 0x20 == 0x20:
      res.append(('priority',self.priority))
    if patch_flags & 0x40 == 0x40:
      res.append(('startup',self.startup))
    if patch_flags & 0x80 == 0x80:
      res.append(('seg_list_blk',self.seg_list_blk))
    if patch_flags & 0x100 == 0x100:
      res.append(('global_vec',self.global_vec))
    return res

  def set_flags(self, flags):
    mask = 0
    for f in flags:
      key = f[0]
      val = int(f[1])
      if key == 'type':
        self.type = val
        mask |= 0x01
      elif key == 'task':
        self.task = val
        mask |= 0x02
      elif key == 'lock':
        self.lock = val
        mask |= 0x04
      elif key == 'handler':
        self.handler = val
        mask |= 0x08
      elif key == 'stack_size':
        self.stack_size = val
        mask |= 0x10
      elif key == 'priority':
        self.priority = val
        mask |= 0x20
      elif key == 'startup':
        self.startup = val
        mask |= 0x40
      elif key == 'seg_list_blk':
        self.seg_list_blk = val
        mask |= 0x80
      elif key == 'global_vec':
        self.global_vec = val
        mask |= 0x100
      else:
        raise ValueError("Invalid flag: "+key)
    return mask

  def get_valid_flag_names(self):
    return ('type', 'task', 'lock', 'handler', 'stack_size', 'priority', 'startup', 'seg_list_blk', 'global_vec')

  def read(self, blk):
    self.type = blk._get_long(11)
    self.task = blk._get_long(12)
    self.lock = blk._get_long(13)
    self.handler = blk._get_long(14)
    self.stack_size = blk._get_long(15)
    self.priority = blk._get_long(16)
    self.startup = blk._get_long(17)
    self.seg_list_blk = blk._get_long(18)
    self.global_vec = blk._get_long(19)

  def write(self, blk):
    blk._put_long(11, self.type)
    blk._put_long(12, self.task)
    blk._put_long(13, self.lock)
    blk._put_long(14, self.handler)
    blk._put_long(15, self.stack_size)
    blk._put_long(16, self.priority)
    blk._put_long(17, self.startup)
    blk._put_long(18, self.seg_list_blk)
    blk._put_long(19, self.global_vec)

class FSHeaderBlock(Block):
  def __init__(self, blkdev, blk_num):
    Block.__init__(self, blkdev, blk_num, chk_loc=2, is_type=Block.FSHD)

  def create(self, host_id=0, next=Block.no_blk, flags=0, dos_type=0, version=0, patch_flags=0,
             size=64, dev_node=None):
    Block.create(self)
    self.size = size
    self.host_id = host_id
    self.next = next
    self.flags = flags

    self.dos_type = dos_type
    self.version = version
    self.patch_flags = patch_flags

    if dev_node == None:
      dev_node = FSHeaderDeviceNode()
    self.dev_node = dev_node

  def write(self):
    self._create_data()

    self._put_long(1, self.size)
    self._put_long(3, self.host_id)
    self._put_long(4, self.next)
    self._put_long(5, self.flags)

    self._put_long(8, self.dos_type)
    self._put_long(9, self.version)
    self._put_long(10, self.patch_flags)

    self.dev_node.write(self)

    Block.write(self)

  def read(self):
    Block.read(self)
    if not self.valid:
      return False

    self.size = self._get_long(1)
    self.host_id = self._get_long(3)
    self.next = self._get_long(4)
    self.flags = self._get_long(5)

    self.dos_type = self._get_long(8)
    self.version = self._get_long(9)
    self.patch_flags = self._get_long(10)

    self.dev_node = FSHeaderDeviceNode()
    self.dev_node.read(self)

    return self.valid

  def get_version_tuple(self):
    return ((self.version >> 16),(self.version & 0xffff))

  def get_version_string(self):
    return "%d.%d" % self.get_version_tuple()

  def get_flags(self):
    return self.dev_node.get_flags(self.patch_flags)

  def get_valid_flag_names(self):
    return self.dev_node.get_valid_flag_names()

  def dump(self):
    Block.dump(self, "FSHeader")

    print(" size:           %d" % self.size)
    print(" host_id:        %d" % self.host_id)
    print(" next:           %s" % self._dump_ptr(self.next))
    print(" flags:          0x%08x" % self.flags)
    print(" dos_type:       0x%08x = %s" % (self.dos_type, DosType.num_to_tag_str(self.dos_type)))
    print(" version:        0x%08x = %s" % (self.version, self.get_version_string()))
    print(" patch_flags:    0x%08x" % self.patch_flags)

    self.dev_node.dump()

  def set_flags(self, flags):
    mask = self.dev_node.set_flags(flags)
    self.patch_flags |= mask

  def set_flag(self, key, value):
    mask = self.dev_node.set_flags([(key,value)])
    self.patch_flags |= mask



import sys
import time

from arcade.glui.dialog import Dialog
from OpenGL import GL as gl

from arcade.glui.render import Render
from arcade.glui.state import State
from arcade.resources import resources

MAX_LINE = 30


def show_exception():
    import traceback

    backtrace = traceback.format_exc()
    State.get().dialog = ErrorDialog(sys.exc_info()[1], backtrace)
    show_error_state = {"stop": False}
    while not show_error_state["stop"]:

        def input_func(button):
            if button == "BACK":
                show_error_state["stop"] = True

        # FIXME
        from arcade.glui.window import main_loop_iteration

        if main_loop_iteration(input_func=input_func):
            break
    State.get().dialog.destroy()
    State.get().dialog = None


class ErrorDialog(Dialog):
    def __init__(self, message, backtrace=None):
        Dialog.__init__(self)
        self.width = 16 / 9 * 2
        self.height = 2.0
        self.message = message
        self.backtrace = backtrace
        self.splitted = self.backtrace.split("\n")
        if not self.splitted[-1]:
            self.splitted = self.splitted[:-1]

        self.background_color = (0.0, 0.0, 0.0, 1.0)
        liberation_mono_bold = resources.resource_filename(
            "LiberationMono-Regular.ttf"
        )
        self.detail_font = pygame.font.Font(
            liberation_mono_bold, int(0.021 * Render.get().display_height)
        )
        self.guru_font = pygame.font.Font(
            liberation_mono_bold, int(0.03 * Render.get().display_height)
        )
        self.start_time = time.time()

    def render_content(self):
        Render.get().dirty = True

        # #x1 = -16 / 9 + 0.1
        # x1 = 0.1
        # #x2 = 16 / 9 - 0.1
        # x2 = self.width - 0.1
        # #y1 = 0.7
        # #y2 = 0.9
        # y1 = 1.6
        # y2 = 1.9
        x1 = 0
        x2 = self.width
        y1 = 1.7
        y2 = 2.0
        w = 0.03

        # t = (pygame.time.get_ticks() - self.start_time) // 1000
        alert_color = (1.0, 0.8, 0.0)
        t = int((time.time() - self.start_time * 1.6))
        if t % 2 == 0:
            gl.glBegin(gl.GL_QUADS)
            gl.glColor3f(*alert_color)
            gl.glVertex2f(x1, y1)
            gl.glVertex2f(x2, y1)
            gl.glVertex2f(x2, y2)
            gl.glVertex2f(x1, y2)
            gl.glColor3f(0.0, 0.0, 0.0)
            gl.glVertex2f(x1 + w, y1 + w)
            gl.glVertex2f(x2 - w, y1 + w)
            gl.glVertex2f(x2 - w, y2 - w)
            gl.glVertex2f(x1 + w, y2 - w)
            gl.glEnd()

        text = "Software Failure.  Press BACKSPACE or back button to continue."
        Render.get().text(text, self.guru_font, 0.2, 1.85, color=alert_color)
        text = self.splitted[-1]
        text = "Guru Meditation #{0}".format(text)
        Render.get().text(text, self.guru_font, 0.2, 1.77, color=alert_color)

        x = 0.2
        y = 0.15

        tw, th = Render.get().measure_text("M", self.detail_font)
        y += th
        lines = []
        max_line_size = 129
        for line in self.splitted:
            line = line.rstrip()
            while len(line) > max_line_size:
                lines.append(line[:max_line_size])
                line = line[max_line_size:]
            lines.append(line)

        for i, line in enumerate(reversed(lines)):
            if i == MAX_LINE:
                break
            s = (MAX_LINE - i) / MAX_LINE
            tw, th = Render.get().text(
                line, self.detail_font, x, y, color=(s, s, s, 1.0)
            )
            y += th



import fsui
from launcher.i18n import gettext
from launcher.option import Option
from launcher.settings.settings_page import SettingsPage


class AdvancedVideoSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("video-settings", "pkg:workspace")
        gettext("Advanced Video Settings")
        title = gettext("Advanced Video")
        subtitle = ""
        self.add_header(icon, title, subtitle)

        self.add_option(Option.FULLSCREEN_MODE)
        self.add_option(Option.VIDEO_FORMAT)
        self.low_latency_group = self.add_option(Option.LOW_LATENCY_VSYNC)

        self.add_section(gettext("OpenGL Settings"))
        self.add_option(Option.FSAA)
        self.add_option(Option.TEXTURE_FILTER)
        self.add_option(Option.TEXTURE_FORMAT)

        # self.add_section(gettext("Video Synchronization"))
        self.sync_method_label = fsui.MultiLineLabel(
            self,
            gettext(
                "Depending on your OS and OpenGL drivers, video synchronization "
                "can use needlessly much CPU (esp. applies to "
                "Linux). You can experiment with different sync methods "
                "to improve performance."
            ),
            640,
        )
        self.layout.add(self.sync_method_label, fill=True, margin_top=20)
        self.sync_method_group = self.add_option(Option.VIDEO_SYNC_METHOD)



from __future__ import absolute_import
from __future__ import print_function

from .ProtectFlags import *
from .TimeStamp import *
from .FSString import FSString

class MetaInfo:
  def __init__(self, protect=None, mod_ts=None, comment=None, protect_flags=None):
    if protect_flags != None:
      self.set_protect_flags(protect_flags)
    else:
      self.set_protect(protect)
    self.set_mod_ts(mod_ts)
    self.set_comment(comment)
  
  def get_str_line(self):
    """Return a unicode string with protect flags, mod time and (optional) comment"""
    res = []
    res.append(self.get_protect_str())
    res.append(self.get_mod_time_str())
    comment = self.get_comment()
    if comment == None:
      res.append(u'')
    else:
      res.append(self.get_comment().get_unicode())
    return u'  '.join(res)
    
  def get_mod_time_str(self):
    if self.mod_ts != None:
      return str(self.mod_ts)
    else:
      return ts_empty_string
      
  def get_protect_str(self):
    if self.protect_flags != None:
      return str(self.protect_flags)
    else:
      return ProtectFlags.empty_string
      
  def get_protect_short_str(self):
    if self.protect_flags != None:
      return self.protect_flags.short_str()
    else:
      return ""
      
  def set_protect(self, protect):
    self.protect = protect
    if self.protect != None:
      self.protect_flags = ProtectFlags(protect)
    else:
      self.protect_flags = None
  
  def set_protect_flags(self, pf):
    self.protect_flags = pf
    self.protect = pf.mask
  
  def set_default_protect(self):
    self.protect = 0
    self.protect_flags = ProtectFlags(self.protect)
  
  def set_current_as_mod_time(self):
    mod_time = time.mktime(time.localtime())
    self.set_mod_time(mod_time)
    
  def set_mod_time(self, mod_time):
    self.mod_time = mod_time
    if self.mod_time != None:
      self.mod_ts = TimeStamp()
      self.mod_ts.from_secs(mod_time)
    else:
      self.mod_ts = None
  
  def set_mod_ts(self, mod_ts):
    self.mod_ts = mod_ts
    if self.mod_ts != None:
      self.mod_time = self.mod_ts.get_secsf()
    else:
      self.mod_time = None
  
  def set_comment(self, comment):
    """Set comment as a FSString"""
    if comment != None and not isinstance(comment, FSString):
      raise ValueError("Comment must be a FSString")
    self.comment = comment
  
  def get_protect(self):
    return self.protect
  
  def get_protect_flags(self):
    return self.protect_flags
  
  def get_mod_time(self):
    return self.mod_time

  def get_mod_ts(self):
    return self.mod_ts

  def get_comment(self):
    return self.comment

  def get_comment_ami_str(self):
    if self.comment != None:
      return self.comment.get_ami_str()
    else:
      return ""
  
  def get_comment_unicode_str(self):
    if self.comment != None:
      return self.comment.get_unicode()
    else:
      return u""



import hashlib
import os

from fsbc import settings
from fsgs.drivers.mednafendriver import MednafenDriver
from fsgs.drivers.mess.messsmsdriver import MessSmsDriver
from fsgs.option import Option
from fsgs.platform import Platform
from fsgs.platforms.loader import SimpleLoader

"""
FIXME: Some PAL games are 256x240, e.g. Fantastic Dizzy
"""


class MasterSystemPlatform(Platform):
    PLATFORM_NAME = "Master System"

    def driver(self, fsgc):
        if settings.get(Option.SMS_DRIVER) == "mess":
            return MessSmsDriver(fsgc)
        else:
            return MasterSystemMednafenFSDriver(fsgc)

    def loader(self, fsgc):
        return MasterSystemLoader(fsgc)


class MasterSystemLoader(SimpleLoader):
    pass


class MasterSystemMednafenDriver(MednafenDriver):
    CONTROLLER = {
        "type": "gamepad",
        "description": "Gamepad",
        "mapping_name": "mastersystem",
    }

    PORTS = [
        {"description": "1st Controller", "types": [CONTROLLER]},
        {"description": "2nd Controller", "types": [CONTROLLER]},
    ]

    def __init__(self, fsgc, vanilla=True):
        super().__init__(fsgc, vanilla)
        self.helper = MasterSystemHelper(self.options)
        # FIXME: Not checked
        self.save_handler.set_save_data_is_emulator_specific(True)
        # self.save_handler.set_srm_alias(".sav")

    def prepare(self):
        print("[SMS] Mednafen driver preparing...")
        super().prepare()
        rom_path = self.helper.prepare_rom(self)
        # self.helper.read_rom_metadata(rom_path)
        # self.init_mega_drive_model()
        # for i in range(len(self.PORTS)):
        #     self.init_mega_drive_port(i)
        # self.init_mednafen_crop_from_viewport()
        self.set_mednafen_aspect(4, 3)

        # We do aspect calculation separately. Must not be done twice.
        # FIXME: Option does not exist, check that end result is correct!
        # self.emulator.args.extend(["-sms.correct_aspect", "0"])

        # ROM path must be added at the end of the argument list
        self.emulator.args.append(rom_path)

    def mednafen_system_prefix(self):
        return "sms"

    def mednafen_input_mapping(self, port):
        if port == 0:
            return {
                "1": "sms.input.port1.gamepad.fire1",
                "2": "sms.input.port1.gamepad.fire2",
                "UP": "sms.input.port1.gamepad.up",
                "DOWN": "sms.input.port1.gamepad.down",
                "LEFT": "sms.input.port1.gamepad.left",
                "RIGHT": "sms.input.port1.gamepad.right",
                "PAUSE": "sms.input.port1.gamepad.pause",
            }
        elif port == 1:
            return {
                "1": "sms.input.port2.gamepad.fire1",
                "2": "sms.input.port2.gamepad.fire2",
                "UP": "sms.input.port2.gamepad.up",
                "DOWN": "sms.input.port2.gamepad.down",
                "LEFT": "sms.input.port2.gamepad.left",
                "RIGHT": "sms.input.port2.gamepad.right",
                "PAUSE": "sms.input.port2.gamepad.pause",
            }

        # FIXME: add PAUSE button to universal gamepad config

    # def game_video_par(self):
    #     size = self.game_video_size()
    #     return (4 / 3) / (size[0] / size[1])

    def game_video_size(self):
        if self.is_pal():
            size = (256, 240)
        else:
            size = (256, 192)
        return size

    def get_game_file(self, config_key="cartridge_slot"):
        return None


class MasterSystemMednafenFSDriver(MasterSystemMednafenDriver):
    def __init__(self, fsgs):
        super().__init__(fsgs, vanilla=False)


class MasterSystemHelper:
    def __init__(self, options):
        self.options = options

    # def model(self):
    #     if self.options[Option.SMD_MODEL] == SMD_MODEL_NTSC:
    #         return SMD_MODEL_NTSC
    #     if self.options[Option.SMD_MODEL] == SMD_MODEL_NTSC_J:
    #         return SMD_MODEL_NTSC_J
    #     if self.options[Option.SMD_MODEL] == SMD_MODEL_PAL:
    #         return SMD_MODEL_PAL
    #     # FIXME: REMOVE?
    #     # return SMD_MODEL_AUTO
    #     return SMD_MODEL_NTSC

    # def set_model_name_from_model(self, driver):
    #     model = self.model()
    #     if model == SMD_MODEL_NTSC:
    #         driver.set_model_name("Genesis")
    #     elif model == SMD_MODEL_PAL:
    #         driver.set_model_name("Mega Drive PAL")
    #     elif model == SMD_MODEL_NTSC_J:
    #         driver.set_model_name("Mega Drive NTSC-J")

    # FIXME: Shared, move into common module (find all occurrences)
    def prepare_rom(self, driver):
        file_uri = self.options[Option.CARTRIDGE_SLOT]
        input_stream = driver.fsgc.file.open(file_uri)
        _, ext = os.path.splitext(file_uri)
        return self.prepare_rom_with_stream(driver, input_stream, ext)

    # FIXME: Shared, move into common module (find all occurrences)
    def prepare_rom_with_stream(self, driver, input_stream, ext):
        sha1_obj = hashlib.sha1()
        path = driver.temp_file("rom" + ext).path
        with open(path, "wb") as f:
            while True:
                data = input_stream.read(65536)
                if not data:
                    break
                f.write(data)
                sha1_obj.update(data)
        new_path = os.path.join(
            os.path.dirname(path), sha1_obj.hexdigest()[:8].upper() + ext
        )
        os.rename(path, new_path)
        return new_path



from arcade.glui.state import State


class Navigatable(object):
    def go_left(self, count=1):
        pass

    def go_right(self, count=1):
        pass

    def go_up(self):
        State.get().down_navigatable = self
        State.get().navigatable = State.get().top_menu

    def go_down(self):
        if State.get().down_navigatable:
            State.get().navigatable = State.get().down_navigatable
            State.get().down_navigatable = None

    def activate(self):
        pass

    def render(self):
        pass



INVALID_BOOT_BLOCK = 1
INVALID_ROOT_BLOCK = 2
INVALID_USER_DIR_BLOCK = 3
INVALID_FILE_HEADER_BLOCK = 4
INVALID_FILE_LIST_BLOCK = 5
INVALID_FILE_DATA_BLOCK = 6
NO_FREE_BLOCKS = 7
UNSUPPORTED_DIR_BLOCK = 8
INVALID_FILE_NAME = 9
NAME_ALREADY_EXISTS = 10
INVALID_SEQ_NUM = 11
FILE_LIST_BLOCK_COUNT_MISMATCH = 12
FILE_DATA_BLOCK_COUNT_MISMATCH = 13
INVALID_BITMAP_BLOCK = 14
BITMAP_BLOCK_COUNT_MISMATCH = 15
BITMAP_SIZE_MISMATCH = 16
DELETE_NOT_ALLOWED = 17
INTERNAL_ERROR = 18
INVALID_PROTECT_FORMAT = 19
INVALID_PARENT_DIRECTORY = 20
FILE_NOT_FOUND = 21
INVALID_VOLUME_NAME = 22

error_names = {
  INVALID_BOOT_BLOCK : "Invalid Boot Block",
  INVALID_ROOT_BLOCK : "Invalid Root Block",
  INVALID_USER_DIR_BLOCK : "Invalid UserDir Block",
  INVALID_FILE_HEADER_BLOCK : "Invalid FileHeader Block",
  INVALID_FILE_LIST_BLOCK : "Invalid FileList Block",
  INVALID_FILE_DATA_BLOCK : "Invalid FileData Block",
  NO_FREE_BLOCKS : "No Free Blocks",
  UNSUPPORTED_DIR_BLOCK : "Unsupported Dir Block",
  INVALID_FILE_NAME : "Invalid File Name",
  NAME_ALREADY_EXISTS : "Name already exists",
  INVALID_SEQ_NUM : "Invalid Sequence Number",
  FILE_LIST_BLOCK_COUNT_MISMATCH : "FileList Block Count Mismatch",
  FILE_DATA_BLOCK_COUNT_MISMATCH : "FileData Block Count Mismatch",
  INVALID_BITMAP_BLOCK : "Invalid Bitmap Block",
  BITMAP_BLOCK_COUNT_MISMATCH : "Bitmap Block Count Mismatch",
  BITMAP_SIZE_MISMATCH : "Bitmap Size Mismatch",
  DELETE_NOT_ALLOWED : "Delete Not Allowed",
  INTERNAL_ERROR : "Internal Error",
  INVALID_PROTECT_FORMAT : "Invalid Protect Format",
  INVALID_PARENT_DIRECTORY : "Invalid Parent Directory",
  FILE_NOT_FOUND : "File not found",
  INVALID_VOLUME_NAME : "Invalid volume name"
}

class FSError(Exception):
  def __init__(self, code, node=None, block=None, file_name=None, extra=None):
    self.code = code
    self.node = node
    self.block = block
    self.file_name = file_name
    self.extra = extra
    
  def __str__(self):
    return self.__unicode__().encode("UTF-8")
    
  def __unicode__(self):
    if error_names.has_key(self.code):
      code_str = unicode(error_names[self.code])
    else:
      code_str = u"?"
    srcs = []
    if self.node != None:
      srcs.append(u"node=" + unicode(self.node))
    if self.block != None:
      srcs.append(u"block=" + unicode(self.block))
    if self.file_name != None:
      srcs.append(u"file_name=" + self.file_name.get_unicode())
    if self.extra != None:
      srcs.append(unicode(self.extra))
    return u"%s(%d):%s" % (code_str, self.code, ",".join(srcs))

    


# Automatically generated - do not edit by hand

from fsgs.option import Option as BaseOption


# noinspection PyClassHasNoInit
class Option(BaseOption):
    @staticmethod
    def get(name):
        return options[name]


# noinspection PyPep8Naming
def N_(x):
    return x


options = {
    Option.A2600_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Atari 2600 database"),
        "type": "Boolean",
    },
    Option.A5200_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Atari 5200 database"),
        "type": "Boolean",
    },
    Option.A7800_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Atari 7800 database"),
        "type": "Boolean",
    },
    Option.A7800_MODEL: {
        "default": "ntsc",
        "description": ("Atari 7800 Model"),
        "type": "Choice",
        "values": [("ntsc", "NTSC"), ("pal", "PAL")],
    },
    Option.A7800_PORT_1_TYPE: {
        "default": "joystick",
        "description": ("Atari 7800 Port 1"),
        "type": "Choice",
        "values": [("none", N_("None")), ("joystick", N_("Joystick"))],
    },
    Option.A7800_PORT_2_TYPE: {
        "default": "joystick",
        "description": ("Atari 7800 Port 2"),
        "type": "Choice",
        "values": [("none", N_("None")), ("joystick", N_("Joystick"))],
    },
    Option.ACCELERATOR: {
        "default": "0",
        "description": N_("Accelerator Board"),
        "type": "Choice",
        "values": [
            ("0", N_("None")),
            ("blizzard-1230-iv", "Blizzard 1230 IV"),
            ("blizzard-1240", "Blizzard 1240"),
            ("blizzard-1260", "Blizzard 1260"),
            ("blizzard-ppc", "Blizzard PPC"),
            ("cyberstorm-ppc", "CyberStorm PPC"),
        ],
    },
    Option.ACCELERATOR_MEMORY: {
        "default": "",
        "description": N_("Accelerator RAM"),
        "type": "Choice",
        "values": [
            ("0", "0 MB"),
            ("1024", "1 MB"),
            ("2048", "2 MB"),
            ("4096", "4 MB"),
            ("8192", "8 MB"),
            ("16384", "16 MB"),
            ("32768", "32 MB"),
            ("65536", "64 MB"),
            ("131072", "128 MB"),
            ("262144", "256 MB"),
        ],
    },
    Option.AMIGA_DATABASE: {
        "default": "1",
        "description": ("Enable/disable use of the Amiga game database"),
        "type": "Boolean",
    },
    Option.AMIGA_DRIVER: {
        "default": "fs-uae",
        "description": ("Amiga Game Driver"),
        "type": "Choice",
        "values": [("fs-uae", "FS-UAE")],
    },
    Option.ARCADE_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Arcade game database"),
        "type": "Boolean",
    },
    Option.ARCADE_FULLSCREEN: {
        "default": "1",
        "description": N_("Arcade Fullscreen"),
        "type": "boolean",
    },
    Option.ARCADE_INITIAL_FAVORITES: {
        "default": "0",
        "description": N_("Arcade starts with favorites filter"),
        "type": "Boolean",
    },
    Option.ARCADE_SEARCH: {
        "default": "",
        "description": ("Enable/disable search function in Arcade"),
        "type": "boolean",
    },
    Option.ARCADE_SHUTDOWN: {
        "default": "",
        "description": ("Arcade Shutdown Command"),
        "type": "",
    },
    Option.ARCADE_THEME: {
        "default": "blue",
        "description": N_("Arcade theme"),
        "type": "Choice",
        "values": [("blue", N_("Blue")), ("red", N_("Red"))],
    },
    Option.ATARI_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Atari ST game database"),
        "type": "Boolean",
    },
    Option.ATARI_MODEL: {
        "default": "1040st",
        "description": ("Atari ST Model"),
        "type": "Choice",
        "values": [
            ("520st", "520ST"),
            ("1040st", "1040ST"),
            ("520ste", "520STE"),
            ("1040ste", "1040STE"),
        ],
    },
    Option.AUDIO_BUFFER_TARGET_BYTES: {
        "default": "8192",
        "description": ("Audio buffer target size (bytes)"),
        "type": "integer",
        "min": 1024,
        "max": 32768,
    },
    Option.AUDIO_BUFFER_TARGET_SIZE: {
        "default": "40",
        "description": N_("Audio buffer target size (ms)"),
        "type": "integer",
        "min": 1,
        "max": 100,
    },
    Option.AUDIO_FREQUENCY: {
        "default": "48000",
        "description": N_("Audio output frequency"),
        "type": "choice",
        "values": [
            ("48000", "48000 Hz"),
            ("44100", "44100 Hz"),
            ("22050", "22050 Hz"),
            ("11025", "11025 Hz"),
        ],
    },
    Option.AUTO_LOAD: {
        "default": "1",
        "description": ("Auto-Load Games"),
        "type": "Boolean",
    },
    Option.AUTO_QUIT: {
        "default": "1",
        "description": ("Auto-Quit"),
        "type": "Boolean",
    },
    Option.AUTOMATIC_INPUT_GRAB: {
        "default": "1",
        "description": N_("Grab Input on Click"),
        "type": "boolean",
    },
    Option.BEZEL: {
        "default": "0",
        "description": ("Bezel"),
        "type": "Choice",
        "values": [("1", N_("Show Bezel")), ("0", N_("Hide Bezel"))],
    },
    Option.BLIZZARD_SCSI_KIT: {
        "default": "0",
        "description": N_("Blizzard SCSI Kit"),
        "type": "boolean",
    },
    Option.BORDER: {
        "default": "1",
        "description": ("Border"),
        "type": "Choice",
        "values": [
            ("1", N_("Small Border")),
            ("0", N_("No Border")),
            ("large", N_("Large Border")),
        ],
    },
    Option.BSDSOCKET_LIBRARY: {
        "default": "0",
        "description": ("UAE bsdsocket.library"),
        "type": "boolean",
    },
    Option.BUILTIN_CONFIGS: {
        "default": "1",
        "description": ("Include built-in configurations"),
        "type": "boolean",
    },
    Option.C64_DATABASE: {
        "default": "0",
        "description": (
            "Enable/disable use of the Commodore 64 (C64) database"
        ),
        "type": "Boolean",
    },
    Option.C64_MODEL: {
        "default": "c64c",
        "description": ("Commodore 64 Model"),
        "type": "Choice",
        "values": [
            ("c64c", "C64C"),
            ("c64c/1541-ii", "C64C, 1541-II Floppy Drive"),
            ("c64", "C64"),
        ],
    },
    Option.C64_PALETTE: {
        "default": "vice",
        "description": ("C64 Palette"),
        "type": "Choice",
        "values": [
            ("c64hq", "C64HQ"),
            ("c64s", "C64S"),
            ("ccs64", "CCS64"),
            ("community-colors", "Community"),
            ("deekay", "Deekay"),
            ("frodo", "Frodo"),
            ("godot", "Godot"),
            ("pc64", "PC64"),
            ("pepto-ntsc-sony", "Pepto NTSC Sony"),
            ("pepto-ntsc", "Pepto NTSC"),
            ("pepto-palold", "Pepto PAL-Old"),
            ("pepto-pal", "Pepto PAL"),
            ("ptoing", "Ptoing"),
            ("rgb", "RGB"),
            ("vice", "Vice"),
        ],
    },
    Option.C64_PORT_1_TYPE: {
        "default": "joystick",
        "description": ("C64 Port 1"),
        "type": "Choice",
        "values": [("none", N_("None")), ("joystick", N_("Joystick"))],
    },
    Option.C64_PORT_2_TYPE: {
        "default": "joystick",
        "description": ("C64 Port 2"),
        "type": "Choice",
        "values": [("none", N_("None")), ("joystick", N_("Joystick"))],
    },
    Option.CD32_DATABASE: {
        "default": "1",
        "description": ("Enable/disable use of the CD32 game database"),
        "type": "Boolean",
    },
    Option.CDROM_DRIVE_0_DELAY: {
        "default": "0",
        "description": N_("Delayed CD-ROM Insert"),
        "type": "Boolean",
    },
    Option.CDROM_DRIVE_COUNT: {
        "default": "",
        "description": ("CD-ROM Drive Count"),
        "type": "Choice",
        "values": [("0", "0"), ("1", "1")],
    },
    Option.CDTV_DATABASE: {
        "default": "1",
        "description": ("Enable/disable use of the CDTV game database"),
        "type": "Boolean",
    },
    Option.CHEATS: {
        "default": "0",
        "description": ("Cheats"),
        "type": "Boolean",
    },
    Option.CHIP_MEMORY: {
        "default": "",
        "description": N_("Chip RAM"),
        "type": "Choice",
        "values": [
            ("256", "256 KB"),
            ("512", "512 KB"),
            ("1024", "1 MB"),
            ("1536", "1.5 MB"),
            ("2048", "2 MB"),
            ("4096", "4 MB"),
            ("8192", "8 MB"),
        ],
    },
    Option.CPC_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Amstrad CPC game database"),
        "type": "Boolean",
    },
    Option.CPU: {
        "default": "",
        "description": N_("CPU"),
        "type": "Choice",
        "values": [
            ("68000", "68000"),
            ("68010", "68010"),
            ("68EC020", "68EC020"),
            ("68020", "68020"),
            ("68EC030", "68EC030"),
            ("68030", "68030"),
            ("68EC040", "68EC040"),
            ("68LC040", "68LC040"),
            ("68040-NOMMU", "68040-NOMMU"),
            ("68040", "68040"),
            ("68EC060", "68EC060"),
            ("68LC060", "68LC060"),
            ("68060-NOMMU", "68060-NOMMU"),
            ("68060", "68060"),
        ],
    },
    Option.CPU_IDLE: {
        "default": "",
        "description": (
            "Relax host CPU usage when using fastest-possible CPU"
        ),
        "type": "integer",
        "min": 0,
        "max": 10,
    },
    Option.CROP: {
        "default": "1",
        "description": ("Crop"),
        "type": "Choice",
        "values": [
            ("0", N_("Full Frame")),
            ("1", N_("Crop")),
            ("border", N_("Small Border")),
        ],
    },
    Option.DATABASE_AUTH: {
        "default": "",
        "description": ("Game database authentication"),
        "type": "string",
    },
    Option.DATABASE_EMAIL: {
        "default": "",
        "description": ("Game database email"),
        "type": "string",
    },
    Option.DATABASE_FEATURE: {
        "default": "0",
        "description": ("Enable online database support (requires restart)"),
        "type": "boolean",
    },
    Option.DATABASE_LOCKER: {
        "default": "0",
        "description": ("OpenRetro.org Locker"),
        "type": "Boolean",
    },
    Option.DATABASE_PASSWORD: {
        "default": "",
        "description": ("Game database password"),
        "type": "string",
    },
    Option.DATABASE_SERVER: {
        "default": "openretro.org",
        "description": ("Game Database Server"),
        "type": "string",
    },
    Option.DATABASE_SHOW_ADULT: {
        "default": "0",
        "description": N_("Adult-Themed Games"),
        "type": "boolean",
    },
    Option.DATABASE_SHOW_GAMES: {
        "default": "1",
        "description": N_("Database Games"),
        "type": "choice",
        "values": [
            ("0", N_("All Games")),
            ("1", N_("Available and Downloadable Games")),
            ("2", N_("Available and Auto-Downloadable Games")),
            ("3", N_("Available Games")),
        ],
    },
    Option.DATABASE_SHOW_UNPUBLISHED: {
        "default": "0",
        "description": N_("Unpublished Games"),
        "type": "boolean",
    },
    Option.DATABASE_USERNAME: {
        "default": "",
        "description": ("Game database user name"),
        "type": "string",
    },
    Option.DEVELOPER_MODE: {
        "default": "",
        "description": ("Developer Mode"),
        "type": "",
    },
    Option.DEVICE_ID: {
        "default": "",
        "description": ("Device ID used with OpenRetro authentication"),
        "type": "string",
    },
    Option.DONGLE_TYPE: {
        "default": "0",
        "description": N_("Hardware Dongle"),
        "type": "Choice",
        "values": [
            ("0", N_("None")),
            ("robocop 3", "RoboCop 3"),
            ("leaderboard", "Leader Board"),
            ("b.a.t. ii", "B.A.T. II"),
            ("italy'90 soccer", "Italy Soccer '90"),
            ("dames grand maitre", "Dames Grand-Maître"),
            ("rugby coach", "Rugby Coach"),
            ("cricket captain", "Cricket Captain"),
            ("leviathan", "Leviathan"),
        ],
    },
    Option.DOS_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the DOS game database"),
        "type": "Boolean",
    },
    Option.DOS_EMULATOR: {
        "default": "dosbox-fs",
        "description": ("DOS Emulator"),
        "type": "Choice",
        "values": [
            ("dosbox-fs", "DOSBox-FS"),
            ("dosbox", "DOSBox"),
            ("dosbox-svn", "DOSBox-SVN"),
        ],
    },
    Option.DOSBOX_SBLASTER_IRQ: {
        "default": "7",
        "description": ("Sound Blaster IRQ"),
        "type": "Choice",
    },
    Option.EFFECT: {
        "default": "2x",
        "description": ("Effect"),
        "type": "Choice",
        "values": [
            ("0", N_("No Effect")),
            ("2x", N_("Pixel Doubling")),
            ("hq2x", "HQ2X"),
            ("scale2x", "Scale2X"),
            ("crt", N_("CRT Emulation")),
        ],
    },
    Option.EMULATOR_PID_FILE: {
        "default": "",
        "description": ("Write emulator PID file"),
        "type": "",
    },
    Option.EXPECT_VERSION: {
        "default": "",
        "description": ("Expect Specific FS-UAE Version"),
        "type": "",
    },
    Option.FADE_IN_DURATION: {
        "default": "500",
        "description": ("Fade-in Duration on Start"),
        "type": "",
    },
    Option.FAST_MEMORY: {
        "default": "",
        "description": N_("Fast RAM"),
        "type": "Choice",
        "values": [
            ("0", "0 MB"),
            ("64", "64 KB"),
            ("128", "128 KB"),
            ("256", "256 KB"),
            ("512", "512 KB"),
            ("1024", "1 MB"),
            ("2048", "2 MB"),
            ("4096", "4 MB"),
            ("8192", "8 MB"),
            ("", ""),
        ],
    },
    Option.FLOPPY_DRIVE_COUNT: {
        "default": "",
        "description": ("Floppy Drive Count"),
        "type": "Choice",
        "values": [("0", "0"), ("1", "1"), ("2", "2"), ("3", "3"), ("4", "4")],
    },
    Option.FLOPPY_DRIVE_SPEED: {
        "default": "100",
        "description": N_("Floppy Drive Speed"),
        "type": "Choice",
        "values": [
            ("0", N_("Turbo")),
            ("100", "100%"),
            ("200", "200 %"),
            ("400", "400 %"),
            ("800", "800 %"),
        ],
    },
    Option.FLOPPY_DRIVE_VOLUME: {
        "default": "20",
        "description": N_("Floppy Drive Volume"),
        "type": "integer",
        "min": 0,
        "max": 100,
    },
    Option.FLOPPY_DRIVE_VOLUME_EMPTY: {
        "default": "25",
        "description": N_("Empty Floppy Drive Volume"),
        "type": "integer",
        "min": 0,
        "max": 100,
    },
    Option.FORCE_ASPECT: {
        "default": "",
        "description": ("Force Aspect"),
        "type": "Double",
    },
    Option.FRAME_TIME: {
        "default": "0",
        "description": ("Frame time (ms)"),
        "type": "",
    },
    Option.FREEZER_CARTRIDGE: {
        "default": "0",
        "description": N_("Freezer Cartridge"),
        "type": "Choice",
        "values": [
            ("0", N_("None")),
            ("hrtmon", N_("HRTMon v2.36 (Built-in)")),
            ("action-replay-2", "Action Replay Mk II v2.14"),
            ("action-replay-3", "Action Replay Mk III v3.17"),
        ],
    },
    Option.FSAA: {
        "default": "0",
        "description": N_("Full-scene anti-aliasing (FSAA)"),
        "type": "choice",
        "values": [("0", N_("Off")), ("2", "2x2"), ("4", "4x4"), ("8", "8x8")],
    },
    Option.FULL_KEYBOARD: {
        "default": "0",
        "description": N_("Start with full keyboard emulation"),
        "type": "boolean",
    },
    Option.FULLSCREEN: {
        "default": "0",
        "description": N_("Start FS-UAE in fullscreen mode"),
        "type": "boolean",
    },
    Option.FULLSCREEN_MODE: {
        "default": "desktop",
        "description": N_("Fullscreen Mode"),
        "type": "Choice",
        "values": [
            ("fullscreen", N_("Fullscreen")),
            ("desktop", N_("Fullscreen Desktop")),
            ("window", N_("Fullscreen Window")),
        ],
    },
    Option.G_SYNC: {
        "default": "ignore",
        "description": ("G-Sync"),
        "type": "Choice",
        "values": [("ignore", "Ignore"), ("1", "On"), ("0", "Off (Linux)")],
    },
    Option.GAME_MODE: {
        "default": "1",
        "description": N_("Use GameMode to optimize performance"),
        "type": "Boolean",
    },
    Option.GB_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Game Boy database"),
        "type": "Boolean",
    },
    Option.GB_PORT_1_TYPE: {
        "default": "builtin",
        "description": ("Game Boy Port 1"),
        "type": "Choice",
        "values": [("builtin", N_("Built-in"))],
    },
    Option.GBA_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Game Boy Advance database"),
        "type": "Boolean",
    },
    Option.GBA_GAMMA: {
        "default": "1.3",
        "description": ("Game Boy Advance Gamma Adjustment"),
        "type": "Choice",
        "values": [
            ("1.0", "1.0"),
            ("1.1", "1.1"),
            ("1.2", "1.2"),
            ("1.3", "1.3"),
            ("1.4", "1.4"),
            ("1.5", "1.5"),
            ("1.6", "1.6"),
            ("1.7", "1.7"),
            ("1.8", "1.8"),
            ("1.9", "1.9"),
            ("2.0", "2.0"),
            ("2.1", "2.1"),
            ("2.2", "2.2"),
            ("2.3", "2.3"),
            ("2.4", "2.4"),
            ("2.5", "2.5"),
            ("2.6", "2.6"),
        ],
    },
    Option.GBA_PORT_1_TYPE: {
        "default": "builtin",
        "description": ("Game Boy Advance Port 1"),
        "type": "Choice",
        "values": [("builtin", N_("Built-in"))],
    },
    Option.GBC_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Game Boy Color database"),
        "type": "Boolean",
    },
    Option.GBC_PORT_1_TYPE: {
        "default": "builtin",
        "description": ("Game Boy Color Port 1"),
        "type": "Choice",
        "values": [("builtin", N_("Built-in"))],
    },
    Option.GOVERNOR_WARNING: {
        "default": "1",
        "description": N_("Warn when CPU governor is not 'performance'"),
        "type": "Boolean",
    },
    Option.GRAPHICS_CARD: {
        "default": "none",
        "description": N_("Graphics Card"),
        "type": "Choice",
        "values": [
            ("none", N_("None")),
            ("uaegfx", "UAEGFX"),
            ("uaegfx-z2", "UAEGFX Zorro II"),
            ("uaegfx-z3", "UAEGFX Zorro III"),
            ("picasso-ii", "Picasso II Zorro II"),
            ("picasso-ii+", "Picasso II+ Zorro II"),
            ("picasso-iv", "Picasso IV"),
            ("picasso-iv-z2", "Picasso IV Zorro II"),
            ("picasso-iv-z3", "Picasso IV Zorro III"),
        ],
    },
    Option.GRAPHICS_MEMORY: {
        "default": "",
        "description": N_("Graphics Card RAM"),
        "type": "Choice",
        "values": [
            ("0", "0 MB"),
            ("1024", "1 MB"),
            ("2048", "2 MB"),
            ("4096", "4 MB"),
            ("8192", "8 MB"),
            ("16384", "16 MB"),
            ("32768", "32 MB"),
            ("65536", "64 MB"),
            ("131072", "128 MB"),
            ("262144", "256 MB"),
        ],
    },
    Option.INITIAL_INPUT_GRAB: {
        "default": "",
        "description": N_("Grab Input on FS-UAE Startup"),
        "type": "boolean",
    },
    Option.IRC_NICK: {
        "default": "",
        "description": N_("IRC nickname"),
        "type": "string",
    },
    Option.IRC_SERVER: {
        "default": "irc.fs-uae.net",
        "description": N_("Custom IRC Server"),
        "type": "String",
    },
    Option.JIT_COMPILER: {
        "default": "0",
        "description": N_("JIT Compiler"),
        "type": "Boolean",
    },
    Option.JOYSTICK_PORT_0_AUTOSWITCH: {
        "default": "1",
        "description": N_("Automatic mouse/joystick mode for mouse port"),
        "type": "boolean",
    },
    Option.JOYSTICK_PORT_0_MODE: {
        "default": "mouse",
        "description": ("Joystick Port 0"),
        "type": "",
    },
    Option.JOYSTICK_PORT_1_MODE: {
        "default": "joystick",
        "description": ("Joystick Port 1"),
        "type": "",
    },
    Option.KEEP_ASPECT: {
        "default": "0",
        "description": N_("Keep aspect ratio when scaling (do not stretch)"),
        "type": "boolean",
    },
    Option.KEYBOARD_INPUT_GRAB: {
        "default": "1",
        "description": N_("Grab keyboard when input is grabbed"),
        "type": "boolean",
    },
    Option.KEYBOARD_KEY_BACKSLASH: {
        "default": "action_key_backslash",
        "description": ("Host Key BACKSLASH"),
        "type": "Choice",
        "values": [
            ("action_key_2b", "Amiga Key 0x2B"),
            ("action_key_30", "Amiga Key 0x30"),
            ("action_key_backslash", "Amiga Key Backslash"),
            ("action_key_equals", "Amiga Key Equals"),
        ],
    },
    Option.KEYBOARD_KEY_EQUALS: {
        "default": "action_key_equals",
        "description": ("Host Key EQUALS"),
        "type": "Choice",
        "values": [
            ("action_key_2b", "Amiga Key 0x2B"),
            ("action_key_30", "Amiga Key 0x30"),
            ("action_key_backslash", "Amiga Key Backslash"),
            ("action_key_equals", "Amiga Key Equals"),
        ],
    },
    Option.KEYBOARD_KEY_INSERT: {
        "default": "action_key_2b",
        "description": ("Host Key INSERT"),
        "type": "Choice",
        "values": [
            ("action_key_2b", "Amiga Key 0x2B"),
            ("action_key_30", "Amiga Key 0x30"),
            ("action_key_backslash", "Amiga Key Backslash"),
            ("action_key_equals", "Amiga Key Equals"),
        ],
    },
    Option.KEYBOARD_KEY_LESS: {
        "default": "action_key_30",
        "description": ("Host Key LESS"),
        "type": "Choice",
        "values": [
            ("action_key_2b", "Amiga Key 0x2B"),
            ("action_key_30", "Amiga Key 0x30"),
            ("action_key_backslash", "Amiga Key Backslash"),
            ("action_key_equals", "Amiga Key Equals"),
        ],
    },
    Option.KICKSTART_SETUP: {
        "default": "1",
        "description": (
            "Show kickstart setup page on startup when all ROMs are missing"
        ),
        "type": "boolean",
    },
    Option.LAUNCHER_CLOSE_BUTTONS: {
        "default": "0",
        "description": N_("Include close buttons in dialogs"),
        "type": "Boolean",
    },
    Option.LAUNCHER_CONFIG_FEATURE: {
        "default": "0",
        "description": (
            "Experimental Config Visualization (Requires Restart)"
        ),
        "type": "Boolean",
    },
    Option.LAUNCHER_FONT_SIZE: {
        "default": "",
        "description": N_("Launcher Font Size"),
        "type": "Integer",
        "min": 6,
        "max": 18,
    },
    Option.LAUNCHER_SETUP_WIZARD_FEATURE: {
        "default": "0",
        "description": ("Experimental Setup Wizard (Requires Restart)"),
        "type": "Boolean",
    },
    Option.LAUNCHER_THEME: {
        "default": "fusion",
        "description": N_("Launcher Theme"),
        "type": "Choice",
        "values": [
            ("standard", N_("Standard Qt Theme")),
            ("fusion", "Fusion Auto"),
            ("fusion-plain", "Fusion"),
            ("fusion-adwaita", "Fusion Adwaita"),
            ("fusion-dark", "Fusion Dark"),
            ("fusion-windows10", "Fusion Windows 10"),
        ],
    },
    Option.LOAD_STATE: {
        "default": "",
        "description": ("Load state by number"),
        "type": "integer",
        "min": 1,
        "max": 9,
    },
    Option.LOG_AUTOSCALE: {
        "default": "0",
        "description": N_("Log Autoscale Changes"),
        "type": "boolean",
    },
    Option.LOG_BSDSOCKET: {
        "default": "0",
        "description": "",
        "type": "boolean",
    },
    Option.LOG_FLUSH: {
        "default": "",
        "description": ("Flush log after each log line"),
        "type": "boolean",
    },
    Option.LOG_INPUT: {
        "default": "0",
        "description": N_("Log Input Events"),
        "type": "boolean",
    },
    Option.LOG_QUERY_PLANS: {
        "default": "",
        "description": ("Log database query plans"),
        "type": "",
    },
    Option.LOW_LATENCY_VSYNC: {
        "default": "1",
        "description": N_("Low latency video sync"),
        "type": "boolean",
    },
    Option.MAME_ARTWORK: {
        "default": "0",
        "description": ("Use MAME Artwork"),
        "type": "Boolean",
    },
    Option.MEDNAFEN_AUDIO_BUFFER: {
        "default": "40",
        "description": ("Mednafen audio buffer"),
        "type": "Integer",
        "min": 0,
        "max": 1000,
    },
    Option.MEDNAFEN_AUDIO_DRIVER: {
        "default": "auto",
        "description": ("Mednafen audio driver"),
        "type": "Choice",
        "values": [
            ("auto", "Auto"),
            ("sdl", "SDL"),
            ("alsa", "ALSA"),
            ("mednafen", "Mednafen Default"),
        ],
    },
    Option.MEDNAFEN_DEINTERLACER: {
        "default": "weave",
        "description": ("Mednafen deinterlacer"),
        "type": "Choice",
        "values": [
            ("weave", "weave"),
            ("bob", "bob"),
            ("bob_offset", "bob_offset"),
            ("blend", "blend"),
            ("blend_rg", "blend_rg"),
        ],
    },
    Option.MEDNAFEN_TEMPORAL_BLUR: {
        "default": "0",
        "description": ("Mednafen temporal blur"),
        "type": "Boolean",
    },
    Option.MIDDLE_CLICK_UNGRAB: {
        "default": "1",
        "description": N_("Ungrab Input on Middle Mouse Button"),
        "type": "boolean",
    },
    Option.MIN_FIRST_LINE_NTSC: {
        "default": "21",
        "description": ("First rendered line (NTSC)"),
        "type": "",
    },
    Option.MIN_FIRST_LINE_PAL: {
        "default": "26",
        "description": ("First rendered line (PAL)"),
        "type": "",
    },
    Option.MONITOR: {
        "default": "middle-left",
        "description": N_("Monitor to display FS-UAE on (fullscreen)"),
        "type": "choice",
        "values": [
            ("left", N_("Left")),
            ("middle-left", N_("Middle Left")),
            ("middle-right", N_("Middle Right")),
            ("right", N_("Right")),
        ],
    },
    Option.MOTHERBOARD_RAM: {
        "default": "",
        "description": ("Motherboard RAM"),
        "type": "Choice",
        "values": [
            ("0", "0 MB"),
            ("1024", "1 MB"),
            ("2048", "2 MB"),
            ("4096", "4 MB"),
            ("8192", "8 MB"),
            ("16384", "16 MB"),
            ("32768", "32 MB"),
            ("65536", "64 MB"),
        ],
    },
    Option.MOUSE_SPEED: {
        "default": "100",
        "description": N_("Mouse Speed (%)"),
        "type": "integer",
        "min": 1,
        "max": 500,
    },
    Option.MSX_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the MSX game database"),
        "type": "Boolean",
    },
    Option.N64_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Nintendo 64 database"),
        "type": "Boolean",
    },
    Option.NDS_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Nintendo DS database"),
        "type": "Boolean",
    },
    Option.NEOGEO_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Neo-Geo game database"),
        "type": "Boolean",
    },
    Option.NEOGEO_MODEL: {
        "default": "mvs",
        "description": ("Neo-Geo Model"),
        "type": "Choice",
        "values": [
            ("mvs", "MVS Europe"),
            ("mvs/jp", "MVS Japan"),
            ("mvs/us", "MVS US"),
            ("aes", "AES"),
            ("aes/jp", "AES Japan"),
        ],
    },
    Option.NEOGEO_PORT_1_TYPE: {
        "default": "joystick",
        "description": ("Neo-Geo Port 1"),
        "type": "Choice",
        "values": [("none", N_("None")), ("joystick", N_("Joystick"))],
    },
    Option.NEOGEO_PORT_2_TYPE: {
        "default": "joystick",
        "description": ("Neo-Geo Port 2"),
        "type": "Choice",
        "values": [("none", N_("None")), ("joystick", N_("Joystick"))],
    },
    Option.NES_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Nintendo (NES) database"),
        "type": "Boolean",
    },
    Option.NES_DRIVER: {
        "default": "mednafen",
        "description": ("NES Game Driver"),
        "type": "Choice",
        "values": [
            ("mednafen-fs", "Mednafen-FS"),
            ("higan", "Higan"),
            ("libretro-nestopia", "Nestopia (RetroArch)"),
            ("mednafen", "mednafen"),
            ("mess", "mess"),
        ],
    },
    Option.NES_EMULATOR: {
        "default": "mednafen",
        "description": ("Nintendo Emulator"),
        "type": "Choice",
        "values": [
            ("mednafen", "Mednafen"),
            ("higan", "Higan"),
            ("retroarch-nestopia", "RetroArch Nestopia"),
        ],
    },
    Option.NES_MODEL: {
        "default": "ntsc",
        "description": ("NES Model"),
        "type": "Choice",
        "values": [
            ("ntsc", "NES NTSC"),
            ("pal", "NES PAL"),
            ("ntsc-j", "Famicom"),
        ],
    },
    Option.NES_PORT_1_TYPE: {
        "default": "gamepad",
        "description": ("NES Port 1"),
        "type": "Choice",
        "values": [
            ("none", N_("None")),
            ("gamepad", N_("Gamepad")),
            ("zapper", "Zapper"),
            ("arkanoid", "Arkanoid"),
        ],
    },
    Option.NES_PORT_2_TYPE: {
        "default": "gamepad",
        "description": ("NES Port 2"),
        "type": "Choice",
        "values": [
            ("none", N_("None")),
            ("gamepad", N_("Gamepad")),
            ("zapper", "Zapper"),
            ("arkanoid", "Arkanoid"),
        ],
    },
    Option.NES_PORT_3_TYPE: {
        "default": "none",
        "description": ("NES Port 3"),
        "type": "Choice",
        "values": [("none", N_("None")), ("gamepad", N_("Gamepad"))],
    },
    Option.NES_PORT_4_TYPE: {
        "default": "none",
        "description": ("NES Port 4"),
        "type": "Choice",
        "values": [("none", N_("None")), ("gamepad", N_("Gamepad"))],
    },
    Option.NETPLAY_FEATURE: {
        "default": "0",
        "description": ("Enable experimental net play GUI (requires restart)"),
        "type": "boolean",
    },
    Option.NETPLAY_TAG: {
        "default": "UNK",
        "description": N_("Net play tag (max 3 characters)"),
        "type": "string",
    },
    Option.NETWORK_CARD: {
        "default": "0",
        "description": N_("Network Card"),
        "type": "Choice",
        "values": [("0", N_("None")), ("a2065", "A2065")],
    },
    Option.NGC_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the GameCube database"),
        "type": "Boolean",
    },
    Option.OFFLINE_MODE: {
        "default": "0",
        "description": ("Offline mode"),
        "type": "Boolean",
    },
    Option.PLATFORM: {
        "default": "amiga",
        "description": ("Platform"),
        "type": "Choice",
        "values": [
            ("amiga", "Amiga"),
            ("cpc", "Amstrad CPC"),
            ("a2600", "Atari 2600"),
            ("a5200", "Atari 5200"),
            ("a7800", "Atari 7800"),
            ("atari", "Atari ST"),
            ("arcade", "Arcade"),
            ("cd32", "CD32"),
            ("cdtv", "CDTV"),
            ("c64", "Commodore 64"),
            ("dos", "DOS"),
            ("gb", "Game Boy"),
            ("gba", "Game Boy Advance"),
            ("gbc", "Game Boy Color"),
            ("ngc", "GameCube"),
            ("neogeo", "Neo-Geo"),
            ("nes", "Nintendo"),
            ("n64", "Nintendo 64"),
            ("nds", "Nintendo DS"),
            ("sms", "Master System"),
            ("smd", "Mega Drive"),
            ("psx", "PlayStation"),
            ("snes", "Super Nintendo"),
            ("tg16", "TurboGrafx-16"),
            ("tgcd", "TurboGrafx-CD"),
            ("zxs", "ZX Spectrum"),
        ],
    },
    Option.PLATFORMS_FEATURE: {
        "default": "0",
        "description": N_("Enable Additional Platforms (Requires Restart)"),
        "type": "Boolean",
    },
    Option.PSX_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the PlayStation database"),
        "type": "Boolean",
    },
    Option.PSX_MODEL: {
        "default": "ntsc",
        "description": ("PlayStation Model"),
        "type": "Choice",
        "values": [
            ("ntsc", "NTSC v3.0A ROM BIOS"),
            ("ntsc-j", "NTSC-J v3.0J ROM BIOS"),
            ("pal", "PAL v3.0E ROM BIOS"),
        ],
    },
    Option.PSX_PORT_1_TYPE: {
        "default": "gamepad",
        "description": ("PlayStation Port 1"),
        "type": "Choice",
        "values": [("none", N_("None")), ("gamepad", N_("Gamepad"))],
    },
    Option.PSX_PORT_2_TYPE: {
        "default": "gamepad",
        "description": ("PlayStation Port 1"),
        "type": "Choice",
        "values": [("none", N_("None")), ("gamepad", N_("Gamepad"))],
    },
    Option.PSX_PRELOAD: {
        "default": "0",
        "description": ("Preload PlayStation CD images in RAM"),
        "type": "Boolean",
    },
    Option.QUICK_SETTINGS_OPTIONS: {
        "default": "",
        "description": ("Quick Settings to Display"),
        "type": "String",
    },
    Option.RAW_INPUT: {
        "default": "1",
        "description": N_("Use keyboard raw input (Windows)"),
        "type": "Boolean",
    },
    Option.RELATIVE_PATHS: {
        "default": "",
        "description": ("Relative paths"),
        "type": "",
    },
    Option.RELATIVE_TEMP_FEATURE: {
        "default": "0",
        "description": ("Relative Temporary Directories"),
        "type": "Boolean",
    },
    Option.RETROARCH_AUDIO_BUFFER: {
        "default": "40",
        "description": ("RetroArch Audio Buffer"),
        "type": "Integer",
        "min": 0,
        "max": 1000,
    },
    Option.RTG_SCANLINES: {
        "default": "0",
        "description": N_("Render scan lines in RTG graphics mode"),
        "type": "boolean",
    },
    Option.SAVE_DISK: {
        "default": "1",
        "description": ("Save Disk"),
        "type": "Boolean",
    },
    Option.SCALE: {
        "default": "1",
        "description": ("Scale"),
        "type": "Choice",
        "values": [
            ("1", N_("Max Scaling")),
            ("0", N_("No Scaling")),
            ("integer", N_("Integer Scaling")),
            ("legacy", N_("Legacy Scaling")),
        ],
    },
    Option.SCANLINES: {
        "default": "0",
        "description": N_("Render scan lines"),
        "type": "boolean",
    },
    Option.SGG_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Sega Game Gear database"),
        "type": "Boolean",
    },
    Option.SGG_PORT_1_TYPE: {
        "default": "builtin",
        "description": ("Game Gear Port 1"),
        "type": "Choice",
        "values": [("builtin", N_("Built-in"))],
    },
    Option.SLOW_MEMORY: {
        "default": "",
        "description": N_("Slow RAM"),
        "type": "Choice",
        "values": [
            ("0", "0 MB"),
            ("512", "512 KB"),
            ("1024", "1 MB"),
            ("1536", "1.5 MB"),
            ("1792", "1.8 MB"),
        ],
    },
    Option.SMD_DATABASE: {
        "default": "0",
        "description": (
            "Enable/disable use of the Sega Mega Drive (Genesis) database"
        ),
        "type": "Boolean",
    },
    Option.SMD_EMULATOR: {
        "default": "mednafen-fs",
        "description": ("Mega Drive Emulator"),
        "type": "Choice",
        "values": [
            ("mame-fs", "MAME-FS"),
            ("mednafen", "Mednafen"),
            ("mednafen-fs", "Mednafen-FS"),
            ("retroarch-fs", "RetroArch-FS"),
            ("retroarch-fs/genesisplusgx", "RetroArch-FS: Genesis Plus GX"),
        ],
    },
    Option.SMD_MODEL: {
        "default": "ntsc",
        "description": ("Mega Drive Model"),
        "type": "Choice",
        "values": [
            ("ntsc", "Genesis"),
            ("pal", "Mega Drive PAL"),
            ("ntsc-j", "Mega Drive NTSC-J"),
        ],
    },
    Option.SMD_PORT_1_TYPE: {
        "default": "gamepad",
        "description": ("Mega Drive Port 1"),
        "type": "Choice",
        "values": [
            ("none", N_("None")),
            ("gamepad", "Control Pad"),
            ("gamepad6", N_("Six Button Control Pad")),
        ],
    },
    Option.SMD_PORT_2_TYPE: {
        "default": "gamepad",
        "description": ("Mega Drive Port 2"),
        "type": "Choice",
        "values": [
            ("none", N_("None")),
            ("gamepad", "Control Pad"),
            ("gamepad6", N_("Six Button Control Pad")),
        ],
    },
    Option.SMOOTHING: {
        "default": "auto",
        "description": ("Smoothing"),
        "type": "Choice",
        "values": [
            ("auto", N_("Auto Smoothing")),
            ("0", N_("No Smoothing")),
            ("1", N_("Smoothing")),
        ],
    },
    Option.SMS_DATABASE: {
        "default": "0",
        "description": (
            "Enable/disable use of the Sega Master System database"
        ),
        "type": "Boolean",
    },
    Option.SMS_DRIVER: {
        "default": "mednafen",
        "description": ("Master System Game Driver"),
        "type": "Choice",
        "values": [("mednafen", "mednafen"), ("mess", "mess")],
    },
    Option.SNES_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the Super Nintendo database"),
        "type": "Boolean",
    },
    Option.SNES_PORT_1_TYPE: {
        "default": "gamepad",
        "description": ("SNES Port 1"),
        "type": "Choice",
        "values": [("none", N_("None")), ("gamepad", N_("Gamepad"))],
    },
    Option.SNES_PORT_2_TYPE: {
        "default": "gamepad",
        "description": ("SNES Port 2"),
        "type": "Choice",
        "values": [("none", N_("None")), ("gamepad", N_("Gamepad"))],
    },
    Option.SOUND_CARD: {
        "default": "0",
        "description": N_("Sound Card"),
        "type": "Choice",
        "values": [("0", N_("None")), ("toccata", "Toccata")],
    },
    Option.STEREO_SEPARATION: {
        "default": "70",
        "description": N_("Stereo Separation"),
        "type": "choice",
        "values": [
            ("100", "100%"),
            ("90", "90%"),
            ("80", "80%"),
            ("70", "70%"),
            ("60", "60%"),
            ("50", "50%"),
            ("40", "40%"),
            ("30", "30%"),
            ("20", "20%"),
            ("10", "10%"),
            ("0", "0%"),
        ],
    },
    Option.STRETCH: {
        "default": "aspect",
        "description": ("Stretch"),
        "type": "Choice",
        "values": [
            ("1", N_("Fill Screen")),
            ("aspect", N_("Correct Aspect")),
            ("0", N_("Square Pixels")),
        ],
    },
    Option.SWAP_CTRL_KEYS: {
        "default": "0",
        "description": N_("Swap left and right CTRL keys"),
        "type": "boolean",
    },
    Option.TEXTURE_FILTER: {
        "default": "linear",
        "description": N_("Texture filter"),
        "type": "choice",
        "values": [("nearest", "GL_NEAREST"), ("linear", "GL_LINEAR")],
    },
    Option.TEXTURE_FORMAT: {
        "default": "",
        "description": N_("Video texture format (on the GPU)"),
        "type": "choice",
        "values": [
            ("rgb", "GL_RGB"),
            ("rgb8", "GL_RGB8"),
            ("rgba", "GL_RGBA"),
            ("rgba8", "GL_RGBA8"),
            ("rgb5", "GL_RGB5"),
            ("rgb5_1", "GL_RGB5_1"),
        ],
    },
    Option.TG16_DATABASE: {
        "default": "0",
        "description": (
            "Enable/disable use of the TurboGrafx-16 game database"
        ),
        "type": "Boolean",
    },
    Option.TGCD_DATABASE: {
        "default": "0",
        "description": (
            "Enable/disable use of the TurboGrafx-CD game database"
        ),
        "type": "Boolean",
    },
    Option.TURBO_LOAD: {
        "default": "1",
        "description": ("Turbo Load"),
        "type": "Boolean",
    },
    Option.UAE_A2065: {
        "default": "",
        "description": "uae_a2065",
        "type": "",
        "values": [("slirp", "slirp")],
    },
    Option.UAE_A3000MEM_SIZE: {
        "default": "",
        "description": "Size in megabytes of motherboard fast memory",
        "type": "integer",
        "min": 0,
        "max": 65536,
    },
    Option.UAE_CHIPSET_COMPATIBLE: {
        "default": "",
        "description": "Enable default chipset features for a specific model",
        "type": "choice",
        "values": [
            ("-", "-"),
            ("Generic", "Generic"),
            ("CDTV", "CDTV"),
            ("CD32", "CD32"),
            ("A500", "A500"),
            ("A500+", "A500+"),
            ("A600", "A600"),
            ("A1000", "A1000"),
            ("A1200", "A1200"),
            ("A2000", "A2000"),
            ("A3000", "A3000"),
            ("A3000T", "A3000T"),
            ("A4000", "A4000"),
            ("A4000T", "A4000T"),
        ],
    },
    Option.UAE_CPU_FREQUENCY: {
        "default": "",
        "description": "Specify the frequency of the emulated CPU in cycle-exact modes",
        "type": "float",
        "min": 1.0,
        "max": 100.0,
    },
    Option.UAE_CPU_MULTIPLIER: {
        "default": "",
        "description": "FIXME",
        "type": "integer",
        "min": 0,
        "max": 256,
    },
    Option.UAE_CPU_SPEED: {
        "default": "",
        "description": "Enable/disable fastest possible CPU speed",
        "type": "choice",
        "values": [
            ("real", "Approximate A500/A1200 or cycle-exact"),
            ("max", "Fastest possible"),
        ],
    },
    Option.UAE_CPU_THROTTLE: {
        "default": "",
        "description": "FIXME",
        "type": "float",
        "min": -900.0,
        "max": 5000.0,
    },
    Option.UAE_FASTMEM2_SIZE: {
        "default": "0",
        "description": "Size in MB of Zorro-II Fast RAM (second) expansion board",
        "type": "choice",
        "values": [("0", "0"), ("1", "1"), ("2", "2"), ("4", "4")],
    },
    Option.UAE_FASTMEM_AUTOCONFIG: {
        "default": "1",
        "description": "Autoconfig Z2 Fast RAM",
        "type": "boolean",
    },
    Option.UAE_FASTMEM_SIZE: {
        "default": "0",
        "description": "Size in MB of Zorro-II Fast RAM expansion board",
        "type": "choice",
        "values": [("0", "0"), ("1", "1"), ("2", "2"), ("4", "4"), ("8", "8")],
    },
    Option.UAE_FORCE_0X10000000_Z3: {
        "default": "false",
        "description": "Force Zorro-III address space at 0x10000000",
        "type": "boolean",
    },
    Option.UAE_GFX_LINEMODE: {
        "default": "",
        "description": "Controls how lines are doubled and interlaced modes are handled",
        "type": "choice",
        "values": [
            ("none", "Single / Single"),
            ("double", "Double / Double Frames"),
            ("double2", "Double / Double Fields"),
            ("double3", "Double / Double Fields+"),
            ("scanlines", "Scanlines / Double Frames"),
            ("scanlines2", "Scanlines / Double Fields"),
            ("scanlines3", "Scanlines / Double Fields+"),
            ("scanlines2p", "Double Fields / Double Frames"),
            ("scanlines2p2", "Double Fields / Double Fields"),
            ("scanlines2p3", "Double Fields / Double Fields+"),
            ("scanlines3p", "Double Fields+ / Double Frames"),
            ("scanlines3p2", "Double Fields+ / Double Fields"),
            ("scanlines3p3", "Double Fields+ / Double Fields+"),
        ],
    },
    Option.UAE_MBRESMEM_SIZE: {
        "default": "",
        "description": "Size in megabytes of processor slot fast memory",
        "type": "integer",
        "min": 0,
        "max": 131072,
    },
    Option.UAE_RTC: {
        "default": "auto",
        "description": "Enable a real time clock (RTC) module.",
        "type": "Choice",
        "values": [
            ("auto", "Auto"),
            ("none", "None"),
            ("MSM6242B", "Oki MSM6242B"),
            ("RP5C01A", "Ricoh RP5C01A"),
            ("MSM6242B_A2000", "A2000 MSM6242B"),
        ],
    },
    Option.UAE_SANA2: {
        "default": "false",
        "description": "uae_sana2",
        "type": "BooleanUAE,",
    },
    Option.UAE_SLIRP_IMPLEMENTATION: {
        "default": "auto",
        "description": "Slirp Implementation",
        "type": "Choice",
        "values": [
            ("auto", "auto"),
            ("none", "none"),
            ("builtin", "builtin"),
            ("qemu", "qemu"),
        ],
    },
    Option.UAE_SOUND_OUTPUT: {
        "default": "",
        "description": "Sound emulation",
        "type": "",
        "values": [
            ("none", "Disabled"),
            ("interrupts", "Emulated, No Output"),
            ("exact", "Enabled"),
        ],
    },
    Option.UAE_TOCCATA: {
        "default": "",
        "description": "Toccata Z2 sound card emulation",
        "type": "uaeyesno",
    },
    Option.UAE_Z3CHIPMEM_SIZE: {
        "default": "",
        "description": "",
        "type": "integer",
    },
    Option.UAE_Z3MAPPING: {
        "default": "auto",
        "description": "JIT Direct compatible Z3 memory mapping",
        "type": "choice",
    },
    Option.UAE_Z3MEM2_SIZE: {
        "default": "",
        "description": "",
        "type": "integer",
    },
    Option.UAE_Z3MEM_SIZE: {
        "default": "",
        "description": "Size in MB of Zorro-III Fast RAM expansion board",
        "type": "integer",
    },
    Option.UAEGFX_CARD: {
        "default": "0",
        "description": ("Deprecated: uaegfx_card"),
        "type": "boolean",
    },
    Option.UAEM_WRITE_FLAGS: {
        "default": "1",
        "description": ("Write .uaem metadata files"),
        "type": "flags",
    },
    Option.VICE_AUDIO_DRIVER: {
        "default": "auto",
        "description": ("Vice Audio Driver"),
        "type": "Choice",
        "values": [("auto", N_("Auto")), ("sdl", "SDL")],
    },
    Option.VIDEO_FORMAT: {
        "default": "bgra",
        "description": N_("Video buffer format and color depth"),
        "type": "choice",
        "values": [
            ("bgra", N_("32-bit BGRA")),
            ("rgba", N_("32-bit RGBA")),
            ("rgb565", N_("16-bit")),
        ],
    },
    Option.VIDEO_SYNC: {
        "default": "0",
        "description": N_("Video synchronization"),
        "type": "choice",
        "values": [("1", N_("Auto")), ("0", N_("Off"))],
    },
    Option.VIDEO_SYNC_METHOD: {
        "default": "auto",
        "description": N_("Video synchronization method"),
        "type": "choice",
        "values": [
            ("auto", "Auto"),
            ("swap", "Swap"),
            ("swap-finish", "Swap-Finish"),
            ("finish-swap-finish", "Finish-Swap-Finish"),
            ("finish-sleep-swap-finish", "Finish-Sleep-Swap-Finish"),
            ("sleep-swap-finish", "Sleep-Swap-Finish"),
            ("swap-fence", "Swap-Fence"),
            ("swap-sleep-fence", "Swap-Sleep-Fence"),
        ],
    },
    Option.VOLUME: {
        "default": "100",
        "description": N_("Main Volume"),
        "type": "integer",
        "min": 0,
        "max": 100,
    },
    Option.WARP_MODE: {
        "default": "0",
        "description": ("Start in warp mode"),
        "type": "Boolean",
    },
    Option.WHDLOAD_BOOT_DIR: {
        "default": "",
        "description": ("Custom boot directory for automatic WHDLoad support"),
        "type": "",
    },
    Option.WHDLOAD_MODEL: {
        "default": "auto",
        "description": ("Override WHDLoad Amiga model"),
        "type": "Choice",
        "values": [
            ("0", "No override"),
            ("A1200", "A1200"),
            ("A1200/NONCE", "A1200/NONCE"),
        ],
    },
    Option.WHDLOAD_PRELOAD: {
        "default": "1",
        "description": N_("Override WHDLoad preload option"),
        "type": "Boolean",
    },
    Option.WHDLOAD_QUIT_KEY: {
        "default": "0",
        "description": N_("Override WHDLoad quit key"),
        "type": "Choice",
        "values": [
            ("0", "No override"),
            ("45", "Escape"),
            ("50", "F1"),
            ("51", "F2"),
            ("52", "F3"),
            ("53", "F4"),
            ("54", "F5"),
            ("55", "F6"),
            ("56", "F7"),
            ("57", "F8"),
            ("58", "F9"),
            ("59", "F10"),
        ],
    },
    Option.WHDLOAD_SPLASH_DELAY: {
        "default": "200",
        "description": N_("Override WHDLoad splash delay"),
        "type": "integer",
        "min": -1,
        "max": 500,
    },
    Option.WINDOW_BORDER: {
        "default": "",
        "description": ("Show window border and decorations"),
        "type": "boolean",
    },
    Option.WINDOW_HIDDEN: {
        "default": "0",
        "description": ("Window is hidden on startup"),
        "type": "Boolean",
    },
    Option.WINDOW_MINIMIZED: {
        "default": "0",
        "description": ("Window is minimized on startup"),
        "type": "Boolean",
    },
    Option.ZOOM: {
        "default": "692x540",
        "description": N_("Zoom Amiga display (crop)"),
        "type": "choice",
        "values": [
            ("auto", N_("Auto")),
            ("auto+border", N_("Auto + Border")),
            ("full", N_("Full Frame")),
            ("640x400", "640x400"),
            ("640x400+border", N_("640x400 + Border")),
            ("640x480", "640x480"),
            ("640x480+border", N_("640x480 + Border")),
            ("640x512", "640x512"),
            ("640x512+border", N_("640x512 + Border")),
            ("692x540", "692x540"),
            ("704x540", "704x540"),
            ("704x566", "704x566"),
            ("724x566", "724x566"),
        ],
    },
    Option.ZORRO_III_MEMORY: {
        "default": "",
        "description": N_("Zorro III Fast Memory"),
        "type": "Choice",
        "values": [
            ("0", "0 MB"),
            ("1024", "1 MB"),
            ("2048", "2 MB"),
            ("4096", "4 MB"),
            ("8192", "8 MB"),
            ("16384", "16 MB"),
            ("32768", "32 MB"),
            ("65536", "64 MB"),
            ("131072", "128 MB"),
            ("262144", "256 MB"),
            ("393216", "384 MB"),
            ("524288", "512 MB"),
            ("786432", "768 MB"),
            ("1048576", "1024 MB"),
        ],
    },
    Option.ZXS_DATABASE: {
        "default": "0",
        "description": ("Enable/disable use of the ZX Spectrum database"),
        "type": "Boolean",
    },
    Option.ZXS_DRIVER: {
        "default": "fuse",
        "description": ("ZX Spectrum Game Driver"),
        "type": "Choice",
        "values": [("fuse", "Fuse"), ("mess", "MESS")],
    },
    Option.ZXS_MODEL: {
        "default": "spectrum",
        "description": ("ZX Spectrum Model"),
        "type": "Choice",
        "values": [
            ("spectrum", "Spectrum 48K"),
            ("spectrum/if2", "Spectrum 48K, Interface 2"),
            ("spectrum128", "Spectrum 128"),
            ("spectrum+3", "Spectrum +3"),
        ],
    },
}



from fsgs.drivers.mednafendriver import MednafenDriver


class MednafenLynxDriver(MednafenDriver):
    CONTROLLER = {
        "type": "gamepad",
        "description": "Built-in Controller",
        "mapping_name": "gamegear",
    }

    PORTS = [{"description": "Controller", "types": [CONTROLLER]}]

    def __init__(self, fsgs):
        super().__init__(fsgs)

    def mednafen_system_prefix(self):
        return "lynx"

    def mednafen_input_mapping(self, _):
        return {
            "A": "lynx.input.builtin.gamepad.a",
            "B": "lynx.input.builtin.gamepad.b",
            "UP": "lynx.input.builtin.gamepad.up",
            "DOWN": "lynx.input.builtin.gamepad.down",
            "LEFT": "lynx.input.builtin.gamepad.left",
            "RIGHT": "lynx.input.builtin.gamepad.right",
            "OPTION1": "lynx.input.builtin.gamepad.option_1",
            "OPTION2": "lynx.input.builtin.gamepad.option_2",
            "PAUSE": "lynx.input.builtin.gamepad.pause",
        }



import os
import sys
import platform
import functools

from fsgs.plugins.pluginmanager import PluginManager


@functools.lru_cache()
def find(name):
    return Expansion(name)


@functools.lru_cache()
def cpu_name():
    machine = platform.machine().lower()
    if machine in ["x86_64", "amd64", "i386", "i486", "i585", "i686"]:
        if platform.architecture()[0] == "32bit":
            return "x86"
        if platform.architecture()[0] == "64bit":
            return "x86-64"
    if machine.startswith("power"):
        if platform.architecture()[0] == "32bit":
            return "PPC"
    raise Exception("Unknown CPU")


@functools.lru_cache()
def os_name():
    if sys.platform.startswith("linux"):
        return "Linux"
    if sys.platform == "win32":
        return "Windows"
    if sys.platform == "darwin":
        return "OSX"
    raise Exception("Unknown OS")


class Expansion:
    def __init__(self, name):
        self.path = os.path.join("Workspace", "Expansion", name)
        assert os.path.isdir(self.path)

    @classmethod
    def find(cls, name):
        return find(name)

    @functools.lru_cache()
    def executable(self, name):
        path = self.path
        path = os.path.join(path, os_name(), cpu_name())
        path = os.path.join(path, name)
        if os_name() == "Windows":
            path += ".exe"
        return path


def find_executable(name):
    return PluginManager.instance().find_executable(name)



import fsui


class Book(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        self.page_titles = []
        self.pages = []
        self.current_page = None

    def add_page(self, function, title=""):
        self.page_titles.append(title)
        self.pages.append(function)

    def set_page(self, page):
        try:
            index = page + 0
        except TypeError:
            for i, p in enumerate(self.pages):
                if page == p:
                    index = i
                    break
            else:
                raise Exception("page not found")
        if self.current_page:
            self.current_page.hide()
            self.layout.remove(self.current_page)
        if callable(self.pages[index]):
            page = self.pages[index](self)
            self.pages[index] = page
        else:
            page = self.pages[index]
        self.layout.add(page, fill=True, expand=True)
        self.current_page = page
        page.show()
        if hasattr(page, "on_show"):
            page.on_show()
        self.layout.update()



import weakref


class ContextAware(object):
    def __init__(self, context):
        self.__context = weakref.ref(context)

    @property
    def gsc(self):
        return self.__context()

    @property
    def context(self):
        return self.__context()



from arcade.glui.font import BitmapFont
from arcade.glui.inputmenu import InputMenu
from arcade.glui.menu import Menu
from arcade.glui.navigatable import Navigatable
from arcade.glui.opengl import fs_emu_blending, fs_emu_texturing
from OpenGL import GL as gl

from arcade.glui.render import Render
from arcade.glui.state import State
from arcade.glui.texture import Texture
from arcade.glui.topmenu import GameCenterItem
from arcade.option import Option
from fsgs.context import fsgs
from fsgs.platform import PlatformHandler

GROUP_SPACING = 26
HEADING_TEXT_LEFT = 1920 - 560 + 40 + 20 - 28
ITEM_TEXT_LEFT = 1920 - 560 + 40 + 20
SIDEBAR_START_Y = 1080 - 90 - 50


# FIXME: REMOVE?
class Transition(object):
    def __init__(self):
        self.value = 0.0
        self.start = 0
        self.end = 0
        self.on_finish = None


enter_transition = Transition()
exit_transition = Transition()


def render_wall():  # brightness=1.0):
    # glClearColor(0.1, 0.1, 0.1, 1.0)
    # glClear(GL_DEPTH_BUFFER_BIT);
    Render.get().hd_perspective()
    # fs_emu_ortho();
    # fs_emu_blending(FALSE);
    # fs_emu_texturing(FALSE);
    z = -0.999

    # glPushMatrix()
    # glTranslate(0.0, 0.0, 1.0)

    # transition y-coordinate between floor and wall
    split = 361
    fs_emu_blending(False)
    fs_emu_texturing(False)
    gl.glBegin(gl.GL_QUADS)

    gl.glColor3f(39.0 / 255.0, 44.0 / 255.0, 51.0 / 255.0)
    gl.glVertex3f(0, split, z)
    gl.glVertex3f(1920, split, z)
    color = 0
    gl.glColor3f(color, color, color)
    gl.glVertex3f(1920, 1020, z)
    gl.glVertex3f(0, 1020, z)

    gl.glVertex3f(0, 1020, z)
    gl.glVertex3f(1920, 1020, z)
    gl.glVertex3f(1920, 1080, z)
    gl.glVertex3f(0, 1080, z)

    color = 0
    gl.glColor3f(color, color, color)
    gl.glVertex3f(0, 0, z)
    gl.glVertex3f(1920, 0, z)
    gl.glColor3f(20.0 / 255.0, 22.0 / 255.0, 26.0 / 255.0)
    gl.glVertex3f(1920, split, z)
    gl.glVertex3f(0, split, z)

    gl.glEnd()
    # fs_emu_texturing(True)

    # fs_emu_perspective();
    # double t_x = t0_x + (t1_x - t0_x) * g_menu_transition;
    # double t_y = t0_y + (t1_y - t0_y) * g_menu_transition;
    # double t_z = t0_z + (t1_z - t0_z) * g_menu_transition;
    # double r_a = r0_a + (r1_a - r0_a) * g_menu_transition;
    #
    # glScaled(16.0 / 9.0, 1.0, 1.0);
    # glTranslated(t_x, t_y, t_z);
    # glRotated(r_a, 0.0, 1.0, 0.0);


class GameMenu(Menu):
    def __init__(self, item):
        Menu.__init__(self)

        self.runner = None
        self.platform_handler = None
        self.last_menu_data = None

        self.temp_fix_configs(item)

        self.items.append(item)
        if self.use_game_center_item():
            self.top.left.append(GameCenterItem())
        # self.top.left.append(HomeItem())
        # self.top.left.append(MenuItem(item.title))
        self.top.set_selected_index(
            len(self.top.left) + len(self.top.right) - 1
        )

        self.context = None
        self.controller = None

        self.create_context()
        self.create_controller()
        print("GameMenu.__init__: controller is", self.controller)

        self.config_list = GameConfigList(self.runner, item)
        self.info_panel = GameInfoPanel()

        self.config_list.info_panel = self.info_panel
        # self.config_list.controller = self.controller
        self.info_panel.config_list = self.config_list

        self.navigatable = self.config_list

        # reset transitions
        enter_transition.start = State.get().time
        enter_transition.end = enter_transition.start + 0.2
        exit_transition.start = 0

    # noinspection PyMethodMayBeStatic
    def temp_fix_configs(self, item):
        # from fsgs.Database import Database
        # local_game_database = Database.get_instance()
        # game_database = fsgs.get_game_database()
        #
        # variants = local_game_database.find_game_variants_new(
        #     game_uuid=item.uuid)
        # print(variants)
        #
        # ordered_list = []
        # for variant in variants:
        #
        #     variant["like_rating"], variant["work_rating"] = \
        #         game_database.get_ratings_for_game(variant["uuid"])
        #     variant["personal_rating"], ignored = \
        #         local_game_database.get_ratings_for_game(variant["uuid"])
        #
        #     # user_rating = variant[5] or 0
        #     # global_rating = variant[3] or 0
        #     # user_rating = 0
        #     # global_rating = 0
        #
        #     variant_uuid = variant["uuid"]  # variant[2]
        #     variant_name = variant["name"]  # variant[1]
        #     variant_name = variant_name.replace("\n", " (")
        #     variant_name = variant_name.replace(" \u00b7 ", ", ")
        #     variant_name += ")"
        #     ordered_list.append(
        #         ((1 - bool(variant["have"]),
        #           1000 - variant["personal_rating"],
        #           1000 - variant["like_rating"]),
        #          (variant_uuid, variant_name, variant["database"])))
        # ordered_list.sort()
        # print("ordered variant list:")
        # for variant in ordered_list:
        #     print("-", variant[1][1])
        # item.configurations = [co[1] for co in ordered_list]
        variants = fsgs.get_ordered_game_variants(item.uuid)
        item.configurations = []
        for variant in variants:
            item.configurations.append(
                (
                    variant["uuid"],
                    variant["name"],
                    variant["database"],
                    variant["have"],
                )
            )

    def create_context(self):
        item = self.items[0]
        variant_uuid = item.configurations[0][0]
        database_name = item.configurations[0][2]
        print("\nitem:\n", item.configurations)
        print("\nitem[0]:\n", item.configurations[0])
        print("\n\nvariant_uuid =", variant_uuid, "\n\n")

        # print("configurations: ", item.configurations)
        # configs = sort_configurations(item.configurations)

        # configs = item.configurations
        # config = configs[0]
        # self.context = GameContext.create_for_game(
        #     item.platform, item.name, config)

        values = fsgs.game.set_from_variant_uuid(database_name, variant_uuid)

        # print("")
        # for key in sorted(values.keys()):
        #     print(" * {0} = {1}".format(key, values[key]))
        # print("")

        self.platform_handler = PlatformHandler.create(fsgs.game.platform.id)
        loader = self.platform_handler.get_loader(fsgs)
        fsgs.config.load(loader.load_values(values))

    def create_controller(self):
        # old_controller = self.controller
        # self.controller = GameHandler.create_for_game_context(
        #     self.context)

        self.runner = self.platform_handler.get_runner(fsgs)

        # self.controller = GameRunner()

    def recreate_controller(self):
        # self.create_controller()
        # self.config_list.set_controller(self.controller)
        print("recreate_controller, actually, just recreating game context")
        self.create_context()
        self.controller.context = self.context

    def render(self):
        # print("GameMenu.render")
        enter_transition.value = 1.0
        if enter_transition.start > 0:
            enter_transition.value = (
                State.get().time - enter_transition.start
            ) / (enter_transition.end - enter_transition.start)
            # prevent render from stopping when animating
            Render.get().dirty = True
        # transition goes from 1.0 ... 0.0
        # enter_transition.value = 1.0 - enter_transition.value
        # finished = 0
        if enter_transition.value >= 1.0:
            enter_transition.value = 1.0
            # finished = 1.0
            enter_transition.start = 0
            # enter_transition.on_finish()
            # return

        exit_transition.value = 0.0
        if exit_transition.start > 0:
            exit_transition.value = (
                State.get().time - exit_transition.start
            ) / (exit_transition.end - exit_transition.start)
            # prevent render from stopping when animating
            Render.get().dirty = True
        # transition goes from 1.0 ... 0.0
        exit_transition.value = 1.0 - exit_transition.value
        # finished = 0
        # if exit_transition.value <= 0.0:
        #     exit_transition.value = 0.0
        #     # finished = 1.0
        #     exit_transition.start = 0
        #     exit_transition.on_finish()
        #     return

        last_menu = State.get().history[-2]
        self.last_menu_data = last_menu.render()

        # self.top_menu_transition = transition

        # glClear(GL_DEPTH_BUFFER_BIT)
        # print(transition)
        # glEnable(GL_DEPTH_TEST)
        # glDepthMask(True)

        # render_wall(transition)
        # render_screen(transition)
        # if finished:
        #     # FIXME: defer to later via an event or similar?
        return None

    def render_transparent(self, data):
        # print("GameMenu.render_transparent")
        last_menu = State.get().history[-2]
        last_menu.render_transparent(self.last_menu_data)

        if exit_transition.value < 1.0:
            opacity = 1.0 - exit_transition.value
            opacity *= 1.25
            if opacity > 1.0:
                opacity = 1.0
            Render.get().hd_perspective()
            gl.glDisable(gl.GL_DEPTH_TEST)
            fs_emu_texturing(False)
            fs_emu_blending(True)
            gl.glBegin(gl.GL_QUADS)
            gl.glColor4f(0.0, 0.0, 0.0, opacity)
            gl.glVertex2f(0.0, 0.0)
            gl.glVertex2f(1920.0, 0.0)
            gl.glVertex2f(1920.0, 1020.0)
            gl.glVertex2f(0.0, 1020.0)
            gl.glEnd()
            gl.glEnable(gl.GL_DEPTH_TEST)

        alpha = enter_transition.value * 0.75

        Render.get().hd_perspective()
        gl.glDisable(gl.GL_DEPTH_TEST)
        fs_emu_texturing(False)
        fs_emu_blending(True)
        gl.glBegin(gl.GL_QUADS)
        gl.glColor4f(0.0, 0.0, 0.0, alpha)
        # Covers, left
        gl.glVertex2f(0.0, 366.0)
        gl.glVertex2f(746.0, 366.0)
        gl.glVertex2f(746.0, 1020.0)
        gl.glVertex2f(0.0, 1020.0)
        # Covers, right
        gl.glVertex2f(1920.0 - 746.0, 366.0)
        gl.glVertex2f(1920.0, 366.0)
        gl.glVertex2f(1920.0, 1020.0)
        gl.glVertex2f(1920.0 - 746.0, 1020.0)
        # Bottom
        gl.glColor4f(0.0, 0.0, 0.0, 0.50)
        gl.glVertex2f(0.0, 0.0)
        gl.glVertex2f(1920.0, 0.0)
        gl.glVertex2f(1920.0, 366.0)
        gl.glVertex2f(0.0, 366.0)
        gl.glEnd()
        gl.glEnable(gl.GL_DEPTH_TEST)

        transition = enter_transition.value
        if exit_transition.value < 1.0:
            # transition = exit_transition.value
            transition = 1.0 + 3.0 - exit_transition.value * 3.0
        # print("render, transition = ", transition)
        self.config_list.render(transition)

        if exit_transition.value <= 0.0:
            exit_transition.value = 0.0
            # finished = 1.0
            exit_transition.start = 0
            exit_transition.on_finish()

    def go_up(self):
        print("GameMenu.go_up")

    def go_down(self):
        print("GameMenu.go_down")

    def go_left(self, count=1):
        print("GameMenu.go_left")

    def go_right(self, count=1):
        print("GameMenu.go_right")

    def activate(self):
        print("GameMenu.activate")


class GameConfigList(Navigatable):
    def __init__(self, controller, game_item):
        self.index = 0
        self.runner = None
        self.game_item = game_item
        self.items = []
        self.set_controller(controller)

    def set_controller(self, controller):
        self.items = []
        self.items.append(Option.create_group("Game Control", -2.5))
        play_option = Option.create_play_option()
        self.items.append(play_option)

        self.items.append(Option.create_group("Variant", -2.5))
        config_option = Option.create_config_option()
        config_option.title = self.game_item.configurations[0][1]
        self.items.append(config_option)

        if len(self.game_item.configurations) > 1:
            self.items.append(Option.create_group("Other Variants", -2.5))
            for config in self.game_item.configurations[1:]:
                config_option = Option.create_config_option()
                config_option.title = config[1]
                self.items.append(config_option)

        # self.items = controller.options
        self.index = 0
        # set index to first non-group item
        for i, item in enumerate(self.items):
            if not item.group:
                self.index = i
                break
        self.runner = controller
        # self.controller = controller
        # print("GameConfigList.set_controller", id(self.controller))

    def go_up(self):
        old_index = self.index
        if self.index > 0:
            new_index = self.index - 1
            while new_index > 0 and self.items[new_index].group:
                new_index -= 1
        else:
            new_index = 0
        if new_index != old_index and not self.items[new_index].group:
            self.index = new_index
        else:
            Navigatable.go_up(self)

    def go_down(self):
        old_index = self.index
        if self.index < len(self.items) - 1:
            new_index = self.index + 1
            while (
                new_index < len(self.items) - 1 and self.items[new_index].group
            ):
                new_index += 1
        else:
            new_index = len(self.items) - 1
        if new_index != old_index and not self.items[new_index].group:
            self.index = new_index

    def go_left(self, count=1):
        print("GameConfigList.go_left")
        pass

    def go_right(self, count=1):
        pass

    def activate(self):
        # Hack for variants
        if self.index > 4:
            # Put chosen configuration at the top
            temp_config = self.game_item.configurations[0]
            # temp_config = self.game_item.configurations[self.index - 4 + 1]
            self.game_item.configurations[0] = self.game_item.configurations[
                self.index - 4
            ]
            self.game_item.configurations[self.index - 4] = temp_config
            # Put chosen variant item at the top
            temp = self.items[3]
            self.items[3] = self.items[self.index]
            self.items[self.index] = temp
            # Select play game option
            self.index = 1
            # Hackish, please clean up...
            print("ACTIVE CONFIGURATION:", self.game_item.configurations[0])
            State.get().current_menu.create_context()
            State.get().current_menu.create_controller()
            self.runner = State.get().current_menu.runner
            return
        result = self.items[self.index].activate()
        if result == "PLAY":
            # noinspection PyDecorator
            # @staticmethod
            def show_input():
                # print("Create input menu, controller = ", id(self.controller))
                # new_menu = InputMenu(self.game_item, self.controller)
                # State.get().history.append(new_menu)
                # # FIXME
                # from .window import set_current_menu
                # set_current_menu(new_menu)

                print("Create input menu, runner = ", id(self.runner))
                new_menu = InputMenu(self.game_item, self.runner)
                State.get().history.append(new_menu)
                # FIXME
                from arcade.glui.window import set_current_menu

                set_current_menu(new_menu)

            exit_transition.start = State.get().time
            exit_transition.end = exit_transition.start + 0.5
            exit_transition.on_finish = show_input

    def render(self, transition=1.0):
        Render.get().hd_perspective()

        # w = 560
        # h = 92
        # x = 1920 - w * transition
        # y = 1080 - 60 - h
        # y = 1080 - 60 - h - 44
        # z = -0.9

        # item_top = 1080 - 90 - 50
        # item_left = 1920 - 560 + 40

        dx = Texture.sidebar_background_shadow.w * (1.0 - transition)

        gl.glPushMatrix()
        gl.glTranslate(0.0, 0.0, 0.7)
        # glTranslated((1.0 - transition) * g_tex_sidebar_background->width,
        # 0, 0)
        # fs_emu_set_texture(g_tex_sidebar_background)
        # fs_emu_render_texture_with_size(g_tex_sidebar_background,
        #         1920 - g_tex_sidebar_background->width,
        #         0, g_tex_sidebar_background->width,
        #         g_tex_sidebar_background->height)
        # fs_emu_render_sidebar()
        fs_emu_blending(True)
        fs_emu_texturing(True)
        gl.glDepthMask(False)
        Texture.sidebar_background_shadow.render(
            1920 - Texture.sidebar_background_shadow.w * transition,
            0,
            Texture.sidebar_background_shadow.w,
            Texture.sidebar_background_shadow.h,
        )
        if transition > 1.0:
            padding = 1920
            Texture.sidebar_background.render(
                1920
                - Texture.sidebar_background_shadow.w * transition
                + Texture.sidebar_background_shadow.w,
                0,
                1920 + padding,
                Texture.sidebar_background.h,
            )
        gl.glDepthMask(True)
        gl.glPopMatrix()

        y = SIDEBAR_START_Y
        for i, item in enumerate(self.items):
            if item.group:
                if i > 0:
                    y -= GROUP_SPACING
            selected = i == self.index and State.get().navigatable == self
            # z = 0.71
            fs_emu_texturing(False)
            gl.glDisable(gl.GL_DEPTH_TEST)

            if selected:
                fg_color = [1.0, 1.0, 1.0, 1.0]
                # gl.glBegin(gl.GL_QUADS)
                #
                # gl.glColor3f(0.00, 0x99 / 0xff, 0xcc / 0xff)
                # gl.glVertex3f(x, y - 18, z)
                # gl.glVertex3f(x + w, y - 18, z)
                # gl.glVertex3f(x + w, y + h - 4, z)
                # gl.glVertex3f(x, y + h - 4, z)
                # # glColor3f(0.6, 0.6, 0.6)
                # # glVertex3f(x, y + 4, z)
                # # glVertex3f(x + w, y + 4, z)
                # # glVertex3f(x + w, y + h, z)
                # # glVertex3f(x, y + h, z)
                # gl.glEnd()
                fs_emu_blending(True)
                fs_emu_texturing(True)
                gl.glDepthMask(False)
                Texture.item_background.render(1920 - 540 + 13 + dx, y - 18)
                gl.glDepthMask(True)
            else:
                fg_color = [1.0, 1.0, 1.0, 1.0]

            # glBegin(GL_QUADS)
            # glColor3f(0.4, 0.4, 0.4)
            # glVertex3f(x, y, z)
            # glVertex3f(x + w, y, z)
            # glVertex3f(x + w, y + 4, z)
            # glVertex3f(x, y + 4, z)
            # glEnd()
            # fs_emu_texturing(True)

            if i > 0:
                fg_color[3] *= 0.35

            text = item.title

            # FIXME: REMOVE
            # fs_emu_blending(False)
            # fs_emu_blending(True)

            if item.group:
                BitmapFont.menu_font.render(
                    text,
                    HEADING_TEXT_LEFT + dx,
                    y + 14,
                    r=0.0,
                    g=0x99 / 0xFF,
                    b=0xCC / 0xFF,
                )
                x, _ = BitmapFont.menu_font.measure(text)
                x += HEADING_TEXT_LEFT + 12
                fs_emu_blending(True)
                fs_emu_texturing(True)
                # gl.glDepthMask(False)
                Texture.heading_strip.render(x + dx, y + 14, 1920 - x, 32)
            else:
                BitmapFont.menu_font.render(text, ITEM_TEXT_LEFT + dx, y + 14)
            gl.glEnable(gl.GL_DEPTH_TEST)
            # text = item.title.upper()
            # tw, th = Render.get().text(text, Font.main_path_font,
            #         x + 40, y + 43, color=fg_color)
            # text = item.subtitle.upper()
            # fg_color[3] = fg_color[3] * 0.4
            # text = item.title.upper()
            # tw, th = Render.get().text(text, Font.list_subtitle_font,
            #         x + 40, y + 18, color=fg_color)
            y -= 54

        # Gradually erase menu items when transitioning
        if transition > 1.0:
            gl.glDisable(gl.GL_DEPTH_TEST)
            # tr = (transition - 1.0) / 3.0
            alpha = max(0.0, (transition - 1.0) / 2.0)
            Texture.sidebar_background.render(
                1920 - Texture.sidebar_background_shadow.w * transition + 200,
                0,
                1920,
                Texture.sidebar_background.h,
                opacity=alpha,
            )
            gl.glEnable(gl.GL_DEPTH_TEST)


class GameInfoPanel(Navigatable):
    def __init__(self):
        self.index = 0
        self.position = 0
        self.config_list = []

    def go_up(self):
        print("GameInfoPanel.go_up")
        if self.index == 0 and self.position == 0:
            Navigatable.go_up(self)

    def go_down(self):
        print("GameInfoPanel.go_down")

    def go_left(self, count=1):
        print("GameInfoPanel.go_left")

    def go_right(self, count=1):
        State.get().navigatable = self.config_list

    def render(self):
        z = -0.999
        fs_emu_texturing(False)
        fs_emu_blending(False)
        gl.glBegin(gl.GL_QUADS)
        gl.glColor3f(0.0, 0.0, 0.0)
        # glVertex3f(0, 365, z)
        # glVertex3f(1270, 365, z)
        gl.glVertex3f(0, 400, z)
        gl.glVertex3f(1270, 400, z)
        gl.glColor3f(0.25, 0.25, 0.25)
        gl.glVertex3f(1270, 1080 - 60, z)
        gl.glVertex3f(0, 1080 - 60, z)
        gl.glEnd()
        # fs_emu_texturing(True)



# import os
import tempfile
import zipfile

# noinspection PyUnresolvedReferences
from typing import BinaryIO, Dict

_archive = None  # type: zipfile.ZipFile
_archive_initialized = False
_temp = {}  # type: Dict[str, str]


def app_name() -> str:
    pass


def archive() -> zipfile.ZipFile:
    global _archive, _archive_initialized
    if _archive_initialized:
        return _archive
    archive_name = app_name() + ".dat"
    _archive = zipfile.ZipFile(archive_name, "r")
    return _archive


def reset() -> None:
    global _archive, _archive_initialized
    _archive = None
    _archive_initialized = False


def path_no_extract(name: str) -> str:
    return name


def path(name: str) -> str:
    try:
        return _temp[name]
    except KeyError:
        pass
    try:
        return path_no_extract(name)
    except LookupError:
        s = stream(name)
        fd, p = tempfile.mkstemp(suffix=name)
        with fd:
            fd.write(s.read())
        _temp[name] = p
        return p

    # if p is None:
    #     raise LookupError(name)
    # return p


# def stream(name: str, mode: str="rb") -> BinaryIO:
def stream(name: str) -> BinaryIO:
    # archive().read()
    return archive().open(name)



import os
import struct

from amitools.binfmt.BinImage import *
from amitools.binfmt.Relocate import *
from KickRom import KickRomAccess


class RomEntryRaw:
  def __init__(self, name, data, relocs=None):
    self.name = name
    self.data = data
    self.relocs = relocs

  def get_size(self):
    return len(self.data)

  def get_data(self, addr):
    if self.relocs is None or len(self.relocs) == 0:
      return self.data
    else:
      data = bytearray(self.data)
      self._relocate(data, addr)
      return data

  def _relocate(self, data, addr):
    for pos in self.relocs:
      val = struct.unpack_from(">I", data, pos)
      val += addr
      struct.pack_into(">I", data, pos, val)


class RomEntryBinImg:
  def __init__(self, name, bin_img):
    self.name = name
    self.bin_img = bin_img
    self.relocator = Relocate(bin_img)

  def get_size(self):
    return self.relocator.get_total_size()

  def get_data(self, addr):
    return self.relocator.relocate_one_block(addr)


class RomEntryRomHdr:
  def __init__(self, name, skip, jmp_addr):
    self.name = name
    self.skip = skip
    self.jmp_addr = jmp_addr

  def get_size(self):
    return self.skip + 8

  def get_data(self, addr):
    data = chr(0xff) * self.skip
    hdr = struct.pack(">II", 0x11114ef9, self.jmp_addr)
    return data + hdr


class RomBuilder:
  def __init__(self, size=512, base_addr=0xf80000, fill_byte=0xff):
    self.size = size # in KiB
    self.base_addr = base_addr
    self.fill_byte = fill_byte
    self.size_bytes = size * 1024
    # state
    self.modules = []
    self.rom_off = 0
    self.left_bytes = self.size_bytes
    self.data_off = 0
    self.error = None

  def get_error(self):
    return self.error

  def get_data_offset(self):
    return self.data_off

  def get_rom_offset(self):
    return self.rom_off + self.data_off

  def get_bytes_left(self):
    return self.left_bytes

  def does_fit(self, num_bytes):
    if num_bytes <= self.left_bytes:
      return True
    return False

  def _add_entry(self, entry):
    n = entry.get_size()
    if not self.does_fit(n):
      self.error = "module '%s' does not fit into ROM!" % (entry.name)
      return None
    # add entry
    self.modules.append(entry)
    self.data_off += n
    self.left_bytes -= n
    return entry

  def build_file_list(self, names):
    files = []
    for mod in names:
      # is an index file?
      if mod.endswith('.txt'):
        base_path = os.path.dirname(mod)
        with open(mod, "r") as fh:
          for line in fh:
            name = line.strip()
            if len(name) > 0:
              f = os.path.join(base_path, name)
              files.append(f)
      else:
        files.append(mod)
    return files

  def add_module(self, name, data, relocs=None):
    e = RomEntryRaw(name, data, relocs)
    return self._add_entry(e)

  def add_bin_img(self, name, bin_img):
    e = RomEntryBinImg(name, bin_img)
    return self._add_entry(e)

  def build_rom(self):
    rom_data = bytearray(self.size_bytes)
    # fill in modules
    addr = self.base_addr + self.rom_off
    off = self.rom_off
    for mod in self.modules:
      n = mod.get_size()
      rom_data[off: off+n] = mod.get_data(addr)
      off += n
      addr += n
    # fill empty space
    fill = chr(self.fill_byte)
    while off < self.size_bytes:
      rom_data[off] = fill
      off += 1
    return rom_data


class KickRomBuilder(RomBuilder):
  def __init__(self, size, kickety_split=True, rom_ver=None,
               **kw_args):
    RomBuilder.__init__(self, size, **kw_args)
    self.rom_ver = rom_ver
    # do we need a rom header at 256k border? (the original ROMs do this)
    if size == 512:
      self.kickety_split = kickety_split
      self.split_offset = 0x40000
    else:
      self.kickety_split = False
      self.split_offset = None
    # check size
    if size not in (256,512):
      raise ValueError("KickROM size must be 256 or 512 KiB!")
    # we need a footer
    self.left_bytes -= KickRomAccess.FOOTER_SIZE
    # extra rom header takes 8
    if self.kickety_split:
      self.left_bytes -= KickRomAccess.ROMHDR_SIZE

  def cross_kickety_split(self, num_bytes):
    if self.kickety_split:
      new_off = self.data_off + num_bytes
      return self.data_off < self.split_offset and new_off > self.split_offset
    else:
      return False

  def add_kickety_split(self):
    jump_addr = self.base_addr + 2
    skip = self.split_offset - self.data_off
    e = RomEntryRomHdr("KicketySplit", skip, jump_addr)
    return self._add_entry(e)

  def build_rom(self):
    rom_data = RomBuilder.build_rom(self)
    # add kick sum
    kh = KickRomAccess(rom_data)
    # ensure that first module brought the header
    if not kh.check_header():
      self.error = "First KickROM module does not contain RomHdr!"
      return None
    # write custom rev?
    if self.rom_ver is not None:
      kh.write_rom_ver_rev(self.rom_ver)
    # write missing entries in footer
    kh.write_ext_footer()
    return rom_data


class ExtRomBuilder(RomBuilder):
  def __init__(self, size, rom_ver=None, add_footer=False,
               kick_addr=0xf80000, **kw_args):
    RomBuilder.__init__(self, size, **kw_args)
    # kick addr for jump
    self.kick_addr = kick_addr
    # set ROM version
    if rom_ver is None:
      self.rom_ver = (45,10)
    else:
      self.rom_ver = rom_ver
    # add footer
    self.add_footer = add_footer
    if add_footer:
      self.left_bytes -= KickRomAccess.FOOTER_SIZE
    # account for header
    self.left_bytes -= KickRomAccess.EXT_HEADER_SIZE
    self.rom_off = KickRomAccess.EXT_HEADER_SIZE

  def build_rom(self):
    rom_data = RomBuilder.build_rom(self)
    # write a header
    kh = KickRomAccess(rom_data)
    kh.write_ext_header(self.kick_addr+2, self.rom_ver)
    # write footer
    if self.add_footer:
      kh.write_ext_footer()
    return rom_data



import os
import sqlite3
import time
from binascii import unhexlify

from fsgs.BaseDatabase import BaseDatabase
from fsgs.FSGSDirectories import FSGSDirectories

SENTINEL = "fae7671d-e232-4b71-b179-b3cd45995f92"
VERSION = 5
RESET_VERSION = 5


class File(dict):
    def __nonzero__(self):
        # noinspection PyTypeChecker
        return bool(self["path"])

    def __bool__(self):
        # noinspection PyTypeChecker
        return bool(self["path"])


class FileDatabase(BaseDatabase):
    def __init__(self, sentinel):
        BaseDatabase.__init__(self, sentinel)
        self.last_file_insert = None
        self.last_file_delete = None

    # This (class) map contains information about checksum -> uri locations
    # discovered during runtime. This map is not saved.
    static_files = {}

    @classmethod
    def add_static_file(cls, path, size, sha1):
        file = File()
        file["sha1"] = sha1
        file["path"] = path
        file["size"] = size
        file["mtime"] = None
        cls.static_files[sha1] = file

    @classmethod
    def get_path(cls):
        return os.path.join(FSGSDirectories.databases_dir(), "Files.sqlite")

    def get_version(self):
        return VERSION

    def get_reset_version(self):
        return RESET_VERSION

    @classmethod
    def instance(cls, new=False):
        if new or not hasattr(cls.thread_local, "file_database"):
            cls.thread_local.file_database = cls(cls.SENTINEL)
        return cls.thread_local.file_database

    @classmethod
    def get_instance(cls):
        if not hasattr(cls.thread_local, "file_database"):
            cls.thread_local.file_database = cls(cls.SENTINEL)
        return cls.thread_local.file_database

    def get_file_ids(self):
        cursor = self.internal_cursor()
        cursor.execute("SELECT id FROM file WHERE parent IS NULL")
        return set([row[0] for row in cursor.fetchall()])

    def get_file_hierarchy_ids(self, path):
        path = self.encode_path(path)
        if path.endswith("/"):
            path = path[:-1]
        a_path = path + "\u002f"  # Forward slash
        b_path = path + "\u0030"  # Forward slash + 1
        cursor = self.internal_cursor()
        cursor.execute(
            "SELECT id FROM file WHERE path >= ? AND path < ?",
            (a_path, b_path),
        )
        return set([row[0] for row in cursor.fetchall()])

    def encode_path(self, path):
        # This only works if both path and FSGSDirectories.base_dir (etc) have
        # been normalized with get_real_case.
        path = path.replace("\\", "/")
        base_dir = FSGSDirectories.get_base_dir()
        if path.startswith(base_dir):
            path = path[len(base_dir) :]
            if path.startswith("/"):
                path = path[1:]
            path = "$/" + path
        return path

    def decode_path(self, path):
        if not path or path[0] != "$":
            return path
        base_dir = FSGSDirectories.get_base_dir() + "/"
        if path.startswith("$/"):
            path = base_dir + path[2:]
        return path

    def find_local_roms(self):
        if not self.connection:
            self.init()
        a = "$/Kickstarts/"
        b = "$/Kickstarts" + "\u0030"  # one more than forward slash
        query = "SELECT id, path FROM file WHERE path >= ? AND path < ?"
        cursor = self.internal_cursor()
        cursor.execute(query, (a, b))
        result = {}
        for row in cursor.fetchall():
            result[self.decode_path(row[1])] = row[0]
        return result

    def delete_file(self, id=None, path=None):
        cursor = self.internal_cursor()
        delete_ids = []
        if id is not None:
            delete_ids.append(id)
        if path is not None:
            path = self.encode_path(path)
            cursor.execute("SELECT id FROM file WHERE path = ?", (path,))
            for row in cursor:
                delete_ids.append(row[0])
        for id in delete_ids:
            cursor.execute(
                "DELETE FROM file WHERE id = ? OR parent = ?", (id, id)
            )
        self.last_file_delete = int(time.time())

    def check_sha1(self, sha1):
        if sha1 in self.static_files:
            # FIXME: Is the count necessary? ref query below, or do we
            # only need a True/False result? If so, change query to
            # check for existence only, and change return value to boolean.
            return True
        cursor = self.internal_cursor()
        cursor.execute(
            "SELECT count(*) FROM file WHERE sha1 = ?",
            (sqlite3.Binary(unhexlify(sha1)),),
        )
        return cursor.fetchone()[0]

    def find_file(self, name="", sha1="", path=""):
        cursor = self.internal_cursor()
        if sha1:
            # First we try to find the file from the temporary static
            # file map, in case we've been told about files from plugins or
            # archives.
            try:
                return self.static_files[sha1]
            except KeyError:
                pass
            sha1 = unhexlify(sha1.encode("ASCII"))
            cursor.execute(
                "SELECT id, path, sha1, mtime, size, parent "
                "FROM file WHERE sha1 = ? LIMIT 1",
                (sqlite3.Binary(sha1),),
            )
        elif name:
            # noinspection SpellCheckingInspection
            cursor.execute(
                "SELECT id, path, sha1, mtime, size, parent "
                "FROM file WHERE name = ? COLLATE NOCASE LIMIT 1",
                (name.lower(),),
            )
        else:
            path = self.encode_path(path)
            cursor.execute(
                "SELECT id, path, sha1, mtime, size, parent "
                "FROM file WHERE path = ? LIMIT 1",
                (path,),
            )
        row = cursor.fetchone()
        result = File()
        if row:
            if row[5]:
                # parent
                cursor.execute("SELECT path FROM file WHERE id = ?", (row[5],))
                path = cursor.fetchone()[0] + row[1]
            else:
                path = row[1]
            path = self.decode_path(path)
            result["id"] = row[0]
            result["path"] = path
            result["sha1"] = row[2]
            result["mtime"] = row[3]
            result["size"] = row[4]
        else:
            result["id"] = None
            result["path"] = None
            result["sha1"] = None
            result["mtime"] = None
            result["size"] = None
        return result

    def find_files(self, ext=None):
        cursor = self.internal_cursor()
        query = "SELECT path FROM file WHERE 1 = 1"
        args = []
        if ext is not None:
            # This is used (for now) to look up .fs-uae files, so..
            # we don't want files from archives (sqlite like is case
            # insensitive by default).
            query += " AND parent is NULL AND path LIKE ?"
            args.append("%" + ext)
        cursor.execute(query, args)
        results = []
        for row in cursor:
            data = {"path": self.decode_path(row[0])}
            results.append(data)
        return results

    def add_file(self, path="", sha1=None, mtime=0, size=0, parent=None):
        self.init()
        path = self.encode_path(path)
        cursor = self.internal_cursor()
        print(
            "[FILES] Add file",
            repr(path),
            repr(sha1),
            repr(mtime),
            repr(size),
            repr(parent),
        )
        if sha1:
            sha1 = unhexlify(sha1)
        cursor.execute(
            "INSERT INTO file (path, sha1, mtime, size, parent) "
            "VALUES (?, ?, ?, ?, ?)",
            (path, sqlite3.Binary(sha1), mtime, size, parent),
        )
        self.last_file_insert = int(time.time())
        return cursor.lastrowid

    def get_last_event_stamps(self):
        cursor = self.internal_cursor()
        cursor.execute(
            "SELECT last_file_insert, last_file_delete FROM metadata"
        )
        row = cursor.fetchone()
        result = {"last_file_insert": row[0], "last_file_delete": row[0]}
        return result

    def update_last_event_stamps(self):
        if self.last_file_insert:
            cursor = self.internal_cursor()
            cursor.execute(
                "UPDATE metadata set last_file_insert = ?",
                (self.last_file_insert,),
            )
            self.last_file_insert = None
        if self.last_file_delete:
            cursor = self.internal_cursor()
            cursor.execute(
                "UPDATE metadata set last_file_delete = ?",
                (self.last_file_delete,),
            )
            self.last_file_delete = None

    def clear(self):
        if not self.connection:
            self.init()
        cursor = self.internal_cursor()
        cursor.execute("DELETE FROM file")

    def update_database_to_version_5(self):
        cursor = self.internal_cursor()
        try:
            cursor.execute("SELECT count(*) FROM file")
        except sqlite3.OperationalError:
            cursor.execute(
                """CREATE TABLE file (
                    id INTEGER PRIMARY KEY,
                    sha1 BLOB,
                    path TEXT,
                    size INTEGER,
                    mtime INTEGER,
                    parent INTEGER
                    )"""
            )
            cursor.execute("CREATE INDEX file_sha1 ON file(sha1)")
            cursor.execute("CREATE INDEX file_path ON file(path)")
            cursor.execute("CREATE INDEX file_parent ON file(parent)")

        cursor.execute(
            "ALTER TABLE metadata ADD COLUMN last_file_insert INTEGER"
        )
        cursor.execute(
            "ALTER TABLE metadata ADD COLUMN last_file_delete INTEGER"
        )



from fsui.qt import QCheckBox, Signal
from fsui.qt.signal import Signal, SignalWrapper
from .widget_mixin import WidgetMixin


class CheckBox(QCheckBox, WidgetMixin):
    changed_signal = Signal()

    def __init__(self, parent, text="", check=False):
        # self._widget = QCheckBox(text, parent.get_container())
        QCheckBox.__init__(self, text, parent.get_container())
        self.changed = SignalWrapper(self, "changed")

        self.init_widget(parent)
        if check:
            self.setChecked(True)
        # self._widget.stateChanged.connect(self.__state_changed)
        self.stateChanged.connect(self.__state_changed)

    def __state_changed(self):
        if not self.changed.inhibit:
            self.on_changed()

    def is_checked(self):
        return self.isChecked()

    def check(self, checked=True):
        self.setChecked(checked)

    def on_changed(self):
        self.changed.emit()


class HeadingCheckBox(CheckBox):
    def __init__(self, parent, text=""):
        CheckBox.__init__(self, parent, text)
        font = self.get_font()
        font.set_bold(True)
        self.set_font(font)



import traceback

import fsui
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.launcher_settings import LauncherSettings
from launcher.option import Option
from launcher.ui.behaviors import OptionsBehavior
from launcher.ui.behaviors.platformbehavior import PlatformEnableBehavior
from launcher.ui.HelpButton import HelpButton


class OptionWidgetFactory(object):
    """Factory for creating user interface groups for options."""

    # noinspection PyShadowingBuiltins
    def __init__(self, check=False, help=False, settings=False, label=True):
        self.check = check
        self.help = help
        self.label_spacing = 10
        self.label = label
        if settings:
            self.options = LauncherSettings
        else:
            self.options = LauncherConfig

    # noinspection PyShadowingBuiltins
    def create(
        self,
        parent,
        name,
        text=None,
        check=None,
        help=None,
        platforms=None,
        label=None,
    ):
        option = Option.get(name)
        return create_option_group(
            parent,
            self.options,
            option,
            name,
            option["type"].lower(),
            text if text is not None else gettext(option["description"]),
            check if check is not None else self.check,
            help if help is not None else self.help,
            self.label_spacing,
            platforms=platforms,
            use_label=(label if label is not None else self.label),
        )


class OptionPanel(fsui.Panel):
    def __init__(self, parent, platforms):
        super().__init__(parent)
        if platforms:
            PlatformEnableBehavior(self, platforms)


def create_option_group(
    parent,
    options,
    option,
    key,
    option_type,
    option_text,
    use_checkbox,
    use_help_button,
    label_spacing,
    platforms=None,
    use_label=True,
):
    # group = fsui.Group(parent)
    group = OptionPanel(parent, platforms)
    group.layout = fsui.HorizontalLayout()

    if use_checkbox:
        group.layout.add(
            OptionCheckBox(group, option_text + ":", options, key),
            margin_right=label_spacing,
        )
    elif use_label:
        group.layout.add(
            fsui.Label(group, option_text + ":"), margin_right=label_spacing
        )

    choice_values = []

    if option_type == "boolean":
        group.layout.add_spacer(0, expand=True)
        group.layout.add(
            BooleanChoiceControl(
                group, options, key, use_checkbox=use_checkbox
            )
        )

    elif option_type == "choice":
        group.layout.add_spacer(0, expand=True)
        choices = option["values"]
        # if use_checkbox:
        #     choices,insert(("", gettext("Auto")))
        group.layout.add(
            ChoiceControl(
                group, options, key, choices, use_checkbox=use_checkbox
            )
        )

    elif option_type == "string":

        def on_changed():
            val = text_field.get_text()
            LauncherConfig.set(key, val.strip())

        text_field = fsui.TextField(group)
        # text_field.set_min_width(400)
        text_field.set_text(LauncherConfig.get(key))
        text_field.on_changed = on_changed
        group.layout.add(text_field, expand=True)

    elif (
        option["type"].lower() == "integer"
        and "min" in option
        and "max" in option
    ):
        assert use_checkbox

        spin_ctrl = SpinValueControl(
            group, options, key, option["min"], option["max"]
        )

        # current = LauncherConfig.get(key)
        # current_int = int(option["default"])
        # if current:
        #     try:
        #         current_int = int(current)
        #     except ValueError:
        #         pass
        # current_int = max(option["min"], min(option["max"], current_int))
        # check_box = fsui.CheckBox(group, gettext("Default"))
        # spin_ctrl = fsui.SpinCtrl(group, option["min"],
        #                           option["max"], current_int)
        # if current == "":
        #     check_box.check()
        #     spin_ctrl.disable()

        # def on_checkbox():
        #     if check_box.is_checked():
        #         spin_ctrl.set_value(int(option["default"]))
        #         spin_ctrl.disable()
        #         LauncherConfig.set(key, "")
        #     else:
        #         spin_ctrl.enable()
        #
        # check_box.on_changed = on_checkbox
        #
        # def on_spin():
        #     val = spin_ctrl.get_value()
        #     val = max(option["min"], min(option["max"], val))
        #     LauncherConfig.set(key, str(val))

        # spin_ctrl.on_changed = on_spin
        # group.layout.add(check_box)
        group.layout.add_spacer(0, expand=True)
        group.layout.add(spin_ctrl, margin_left=10)

    if choice_values:

        def on_changed():
            index = choice.get_index()
            LauncherConfig.set(key, choice_values[index][0])

        choice_labels = [x[1] for x in choice_values]
        choice = fsui.Choice(group, choice_labels)
        current = LauncherConfig.get(key)
        for i, value in enumerate(choice_values):
            if current == value[0]:
                choice.set_index(i)
                break
        choice.on_changed = on_changed
        group.layout.add_spacer(0, expand=True)
        group.layout.add(choice)
        group.widget = choice

    if use_help_button:
        help_button = HelpButton(parent, "https://fs-uae.net/options#" + key)
        group.layout.add(help_button, margin_left=10)

    return group


class ConfigWidgetFactory(OptionWidgetFactory):
    """Factory for creating user interface groups for config options."""

    # noinspection PyShadowingBuiltins
    def __init__(self, check=True, help=False, label=True):
        super().__init__(check=check, help=help, label=label)


class SettingsWidgetFactory(OptionWidgetFactory):
    """Factory for creating user interface groups for settings."""

    # noinspection PyShadowingBuiltins
    def __init__(self, check=False, help=True):
        super().__init__(check=check, help=help, settings=True)


def default_value(key):
    option = Option.get(key)
    return option["default"]


class OptionCheckBox(fsui.CheckBox):
    def __init__(self, parent, text, options, key):
        super().__init__(parent, text)
        self.implicit_value = ""
        self.options = options
        self.key = key
        setattr(self, "on_{}_option".format(key), self.on_explicit_option)
        setattr(
            self,
            "on___implicit_{}_option".format(key),
            self.on_implicit_option,
        )
        OptionsBehavior(self, options, [key, "__implicit_" + key])

    def on_explicit_option(self, value):
        if value:
            if not self.is_checked():
                with self.changed.inhibit:
                    self.check(True)
        else:
            if self.is_checked():
                with self.changed.inhibit:
                    self.check(False)

    def on_implicit_option(self, value):
        self.implicit_value = value

    def on_changed(self):
        if self.is_checked():
            if self.options.get(self.key) == "":
                print("-- setting option to default value")
                value = self.implicit_value
                if not value:
                    value = default_value(self.key)
                self.options.set(self.key, value)
        else:
            self.options.set(self.key, "")


class ChoiceControl(fsui.Choice):
    def __init__(self, parent, options, key, choices, use_checkbox=False):
        if not use_checkbox:
            choices.insert(0, ("", gettext("Auto")))
        self.choice_values = [x[0] for x in choices]
        self.choice_labels = [x[1] for x in choices]
        super().__init__(parent, self.choice_labels)
        self.implicit_value = ""
        self.options = options
        self.key = key
        self.use_checkbox = use_checkbox
        self.invalid_index = None

        setattr(self, "on_{}_option".format(key), self.on_explicit_option)
        setattr(
            self,
            "on___implicit_{}_option".format(key),
            self.on_implicit_option,
        )
        OptionsBehavior(self, options, [key, "__implicit_" + key])

    def on_explicit_option(self, value):
        print("ChoiceControl", self.key, value)
        self.set_from_value(value)
        if self.use_checkbox:
            self.set_enabled(bool(value))

    def on_implicit_option(self, value):
        self.implicit_value = value
        if self.use_checkbox:
            self.set_from_value()
        else:
            self.update_auto_item()

    def set_from_value(self, value=None):
        if value is None:
            value = self.options.get(self.key)
        if self.use_checkbox:
            if value == "":
                value = self.implicit_value
                # FIXME: This can be removed if/when implicit value always
                # takes into account the default value.
                if not value:
                    value = default_value(self.key)
        for i, v in enumerate(self.choice_values):
            if v == value:
                with self.changed.inhibit:
                    self.set_index(i)
                if self.invalid_index is not None and i != self.invalid_index:
                    self.remove_invalid_item()
                break
        else:
            print(
                "WARNING: Could not find value", repr(value), "key =", self.key
            )
            self.set_invalid_item(value)

    def set_invalid_item(self, value):
        text = "???"
        if self.invalid_index is None:
            self.invalid_index = self.add_item(text)
            self.choice_values.append("")
            self.choice_labels.append("")
        self.choice_values[-1] = value
        self.choice_labels[-1] = text
        with self.changed.inhibit:
            self.set_index(self.invalid_index)

    def remove_invalid_item(self):
        assert self.invalid_index is not None
        self.remove_item(self.invalid_index)
        del self.choice_values[-1]
        del self.choice_labels[-1]
        self.invalid_index = None

    def on_changed(self):
        index = self.get_index()
        self.options.set(self.key, self.choice_values[index])
        if not self.use_checkbox:
            self.update_auto_item()

    def update_auto_item(self):
        # if self.use_checkbox:
        #     # Checkboxes are used, so there isn't an auto entry!
        #     # return
        #     # if self.options.get(self.key):
        #     #     self.set_from_value(self.implicit_value)
        #     pass
        # else:
        if self.get_index() == 0 and self.implicit_value:
            text = gettext("Auto - {}".format(self.implicit_value))
        else:
            text = gettext("Auto")
        self.set_item_text(0, text)


class BooleanChoiceControl(ChoiceControl):
    def __init__(self, parent, options, key, use_checkbox=False):
        choices = [("1", gettext("Enabled")), ("0", gettext("Disabled"))]
        # if not use_checkbox:
        #     choices.insert(0, ("", gettext("Auto")))
        super().__init__(
            parent, options, key, choices, use_checkbox=use_checkbox
        )


class SpinValueControl(fsui.SpinCtrl):
    # noinspection PyShadowingBuiltins
    def __init__(self, parent, options, key, min, max):
        # value = self.int_value(options.get(key))
        # try:
        #     int_value = int(value)
        # except ValueError:
        #     int_value = self.min

        super().__init__(parent, min, max, min)
        self.min = min
        self.max = max
        self.implicit_value = ""
        self.options = options
        self.key = key

        setattr(self, "on_{}_option".format(key), self.on_explicit_option)
        setattr(
            self,
            "on___implicit_{}_option".format(key),
            self.on_implicit_option,
        )
        OptionsBehavior(self, options, [key, "__implicit_" + key])

    def on_changed(self):
        val = self.get_value()
        val = max(self.min, min(self.max, val))
        self.options.set(self.key, str(val))

    def on_explicit_option(self, value):
        print("SpinValueControl", self.key, value)
        self.set_from_value(value)
        # if self.use_checkbox:
        self.set_enabled(bool(value))

    def on_implicit_option(self, value):
        print("SpinValueControl implicit value", value)
        self.implicit_value = value
        self.set_from_value()

    def int_value(self, value):
        try:
            int_value = int(value)
        except ValueError:
            traceback.print_exc()
            int_value = self.min
        # FIXME: Check range?
        return int_value

    def set_from_value(self, value=None):
        if value is None:
            value = self.options.get(self.key)
        if value == "":
            value = self.implicit_value
            # FIXME: This can be removed if/when implicit value always
            # takes into account the default value.
            if not value:
                value = default_value(self.key)
        with self.changed.inhibit:
            self.set_value(self.int_value(value))


# current = LauncherConfig.get(key)
# current_int = int(option["default"])
# if current:
#     try:
#         current_int = int(current)
#     except ValueError:
#         pass
# current_int = max(option["min"], min(option["max"], current_int))
# check_box = fsui.CheckBox(group, gettext("Default"))
# spin_ctrl = fsui.SpinCtrl(group, option["min"],
#                           option["max"], current_int)
#
# if current == "":
#     check_box.check()
#     spin_ctrl.disable()
#
#
# def on_checkbox():
#     if check_box.is_checked():
#         spin_ctrl.set_value(int(option["default"]))
#         spin_ctrl.disable()
#         LauncherConfig.set(key, "")
#     else:
#         spin_ctrl.enable()
#
#
# check_box.on_changed = on_checkbox
#
#
# def on_spin():
#     val = spin_ctrl.get_value()
#     val = max(option["min"], min(option["max"], val))
#     LauncherConfig.set(key, str(val))
